/***********************************************************************
pftrail.cpp v1.1 alpha 1 (28 April 2020)

Copyright 2020 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file is part of the pftrail package. 
The package includes the following files:

* pftrail.cpp
* ifs-classics
* ifs-inventions
* generate-preamble.cpp
* colours
* postamble
* manual.pdf

New versions of the package may be available from 
http://spacefillingcurves.net or http://herman.haverkort.net.

***********************************************************************/

const char* pftrailInfo = 
"pftrail v1.1 ALPHA 1 by Herman Haverkort.";
 
// uncomment the following line to compile the code sections for the -n option
// to widen narrow passages
// #define NARROWS

// uncomment the following line to compile the code sections for the 3D
// drawing style
// #define PRINTABLE

#ifdef PRINTABLE
#define NARROWS
#endif

#include <string>
using std::string;
#include <string.h>
#include <sstream>
using std::stringstream;
#include <iostream>
using std::istream;
using std::ostream;
using std::cout;
using std::cerr;
using std::endl;
using std::fixed;
#include <fstream>
using std::ifstream;
#include <iomanip>
using std::setprecision;
#include <cstdlib>
#include <utility>
#include <stack>
#include <vector>
#include <algorithm>
using std::min;
using std::max;
using std::swap;
#include <cmath>
using std::abs;

/*********************************************************
* Tunable constants etc.                                 *
*********************************************************/

// for polynomial zoom, the precision of doubles may be limiting:
typedef long double DOUBLE;

// precision of the coordinates in the output:
const int outputPrecision = 5;

// constant used to:
// - not choke on computation-induced rounding errors in computeMinDisk;
// - not choke on computation- or IFS-specification-induced rounding 
//   errors in the IFS parser, more specifically the part that checks
//   whether the IFS indeed generates a curve of fractal dimension
//   two, and which part of the curve fills how much of plane;
// - round coordinates that are probably zero to actually zero when
//   the IFS parser converts the input to Cartesian coordinates:
const DOUBLE practicallyZero = 0.0000001;

// maxMeshFill says with how many vertices (modulo the addition of
// another polygon) a mesh is considered to be full:
const int maxMeshFill = 65000;

// cornerDepth determines how tall/flat the model is;
// its value is half of the ratio between the edge length 
// (or equivalently: the short diagonal) and the height (1)
// of the rendering grid:
const DOUBLE cornerDepth = 1.06066; 

// floorHeight determines the height of cells outside the model and
// the upper reference plane that are taken into account in the 
// "eroded" drawing style. Note that 0 is the height of the lower
// reference plane:
const DOUBLE floorHeight = -0.001;

// the minimum elevation difference for which parapets
// are created:
const DOUBLE noParapetDrop = 0.001;

// the minimum elevation difference for which parapets
// have their maximum height: 
const DOUBLE maxParapetDrop = 0.02;

// the next constant controls the maximum horizontal thickness of
// parapets; the higher the number, the thinner the parapets.
// setting it too large could create artefacts from rounding errors
// in the output
const DOUBLE minSurfaceToParapetRatio = 10;

// the zoom exponent for vertical distance as a function
// of the zoom exponent for horizontal distance:
DOUBLE elevationZoomExponent(DOUBLE zoomExponent)
{
  return max((DOUBLE) 1.0, (zoomExponent * 1.5 - 0.5));
} 

// resolution of progress indicators
const int progressUnits = 20;


/*********************************************************
* Input/output headers                                   *
*********************************************************/

enum ErrorLevel 
{ 
  warning, 
  userError, 
  ifsParsingError, 
  usageError, 
  internalError 
};

class Complaint 
// an object to which an error message can be witten with <<. 
// when the object is destroyed, the error message is produced
// and, if it is not a mere warning, the program halts.
{
    private:
  static const char* errorLabel[];
  stringstream complaint;
  ErrorLevel errorLevel;
    public:
  Complaint(ErrorLevel _errorLevel = userError);
  Complaint(ErrorLevel _errorLevel, const string& expected, const string& read);
  ~Complaint();
  template <typename T> Complaint& operator<<(const T& t);
};

class CommentedInput 
// an object that reads a file and while skipping comments.
// Comments are from // to the end of the line, or 
// between /* and */; in the latter case, /* has to be at
// the beginning of the line.
{
    private:
  ifstream source;
  stringstream buffer;
  void fillBuffer();

    public:
  CommentedInput(string& fileName);
  bool fail() const;
  void clear();
  template < class T > CommentedInput& operator>>(T& t);
  void getLine(string& s);
  void getLowerCaseWord(string& s);

  // The following method scans the input and stops after 
  // reading two consecutive words keyword and name 
  // (case-insensitive). Returns true iff the combination
  // was found. 
  // More precisely: after an occurrence of keyword 
  // the next word is compared to name; if it is not a 
  // match, the word is ignored. So, keyword keyword name
  // is not a match, as the first occurrence of keyword
  // "consumes" the second.
  bool find(const char* keyword, const string& name);
};

void makeLowerCase(string& s);

// convert a string to a number, verifying that the string 
// does not contain any further content:
DOUBLE strtofstop(const char* str);
int strtoistop(const char* str);

void reportProgress(DOUBLE x, string description);

/*********************************************************
* Geometry headers                                       *
*********************************************************/

struct Point
{
  DOUBLE x; DOUBLE y;
  Point(DOUBLE _x = 0, DOUBLE _y = 0);
  Point operator+(const Point& p) const;
  Point& operator+=(const Point& p);
  Point operator-(const Point& p) const;
  Point& operator-=(const Point& p);
  Point operator-() const;
  DOUBLE operator*(const Point& p) const; // inner product
  Point operator*(const DOUBLE d) const;
  Point& operator*=(const DOUBLE d);
  Point operator/(const DOUBLE d) const;
  Point& operator/=(const DOUBLE d);
  DOUBLE sqLength() const;
  DOUBLE length() const;
};
ostream& operator<<(ostream& s, const Point& p);

struct ElevatedPoint: public Point
{
  DOUBLE h;
  ElevatedPoint();
  ElevatedPoint(const Point& _p, DOUBLE _h);
  ElevatedPoint(DOUBLE _x, DOUBLE _y, DOUBLE _h);
  ElevatedPoint operator+(const ElevatedPoint& e) const;
  ElevatedPoint& operator+=(const Point& p);
  ElevatedPoint& operator+=(const ElevatedPoint& e);
  ElevatedPoint operator-(const ElevatedPoint& e) const;
  ElevatedPoint& operator-=(const Point& p);
  ElevatedPoint& operator-=(const ElevatedPoint& e);
  ElevatedPoint operator*(const DOUBLE d) const;
  ElevatedPoint operator/(const DOUBLE d) const;
};
ostream& operator<<(ostream& s, const ElevatedPoint& p);

enum SampleCategory 
{ 
  sampleFill, 
  sampleJump, 
  sampleJumpCore
};

struct AnnotatedPoint: public ElevatedPoint
{
  SampleCategory a;
  std::pair< DOUBLE, DOUBLE > range;
  AnnotatedPoint();
  AnnotatedPoint(const ElevatedPoint& _e, SampleCategory _a, DOUBLE r = 0);
  AnnotatedPoint(const Point& _p, DOUBLE _h, SampleCategory _a, DOUBLE r = 0);
};

// to compute the radius of a smallest enclosing circle:
DOUBLE radius(const std::vector< Point >& points);

struct Rotation
// a rotation with scaling in 2D, defaults to identity transformation
{
  Point x; Point y; // corresponding rows of the rotation matrix
  Rotation(const Point& _x = Point(1,0), const Point& _y = Point(0,1));
  Rotation transpose() const;

  // composition *this after r:
  Rotation operator*(const Rotation& r) const; 
  Rotation operator*=(const Rotation& r);

  // apply to point:
  Point operator*(const Point& p) const;
  Point operator()(const Point& p) const;

  // (squared) scaling factor, assuming the transformation is indeed
  // a similarity transformation (not just any linear transformation):       
  DOUBLE sqScale() const; 
  DOUBLE scale() const;
  int orientation() const;
};

struct Similarity
// similarity transformations that consist of rotation, scaling, 
// translation and/or reversal (changing the sign of the 3rd coordinate);
// these can be understood as a 4x4 matrix of which:
// - the upper left quadrant encodes rotation and scaling of x, y;
// - the last elements of the first rows encode translation of x, y, t
// - the third element of the third row encodes scaling/reversal of t
// - the fourth element of the fourth row is 1
// - all other elements are zero.
{
  Rotation rotation;         // upper left quadrant of the 4x4 matrix
  ElevatedPoint translation; // last elements of first three rows
  DOUBLE scale;              // third element of third row
  Similarity(
    const Rotation& _rotation = Rotation(), 
    const ElevatedPoint& _translation = ElevatedPoint(), 
    DOUBLE _scale = 1
  );

  // composition *this after s:
  Similarity operator*(const Similarity& s) const;
  Similarity& operator*=(const Similarity& s);

  // composition s after *this (store result in *this):
  Similarity& apply(const Similarity& s);
  
  // apply to point:
  Point operator()(const Point& p) const;
  ElevatedPoint operator()(const ElevatedPoint& e) const;

  // print with labels left and right of the first row:
  void print( 
    std::ostream& o, 
    const string& leftLabel, 
    const string& rightLabel = string()
  ) const; 
};

// useful constants
const DOUBLE SQRTTHIRD = sqrt((DOUBLE) 1.0/ (DOUBLE) 3.0);
const DOUBLE SQRTTHREE = sqrt((DOUBLE) 3.0);

template < class T >
class TrianglePoint
// points with coordinates on three axes at 120 degree angles 
// (or equivalent: 3D points that represent their orthogonal 
// projections on the plane x + y + z = 0);
// with conversions from/to 2D Cartesian points 
// note: in the three coordinates' system, lines x = a and x = b
// are a distance b-a apart when measured along the y- or z-axis;
// and a distance sqrt(3/4) (b-a) apart when measured orthogonal
// to the x-axis. 
{
    private:
  T x[3];
  static const Rotation TriToCar;

    public:
  // constructor sets x[s mod 3] = a, x[s+1 mod 3] = b, x[s+2 mod 3] = -(a+b):
  TrianglePoint(int s = 0, T a = 0, T b = 0); 
  TrianglePoint(const Point& p);
  T operator[](int i) const;
  operator Point() const;
};

// to round a TrianglePoint to the nearest integer TrianglePoint:
TrianglePoint< int > round(const TrianglePoint< DOUBLE >& t);


/*********************************************************
* IFS headers (parsing, analysis and sampling)           *
*********************************************************/

class Options;

class PointCollector;

struct Segment 
{
  Similarity transformation;
  int generator;
  Segment(
    Similarity _transformation = Similarity(),
    int _generator = 0
  );
};

struct Generator 
{
  // the plane-filling segments
  std::vector< Segment > segments;

  // indices in segments array (store -1 for jumps)
  std::vector< int > segmentIndex;

  // orientation of all segments (used only for jumps)
  std::vector< int > orientation;

  DOUBLE maxRadius;
  DOUBLE maxSampleLength;
  DOUBLE maxSqSampleLength;
};

enum BridgeStyle 
{ 
  noBridges, 
  pointedBridges, 
  rectangularBridges 
};

class IFS
{
    private:
  std::vector< Generator > generators;
  bool continuous;
  DOUBLE diameter; // lower bound on ratio image diameter : generator length
  DOUBLE curveSampleRadius; // radius of region that is covered by a sample
  BridgeStyle bridgeStyle;
  bool widePassages;
  DOUBLE bridgeSampleDistance;
  DOUBLE bridgeWidthFactor;
  DOUBLE bridgeCoreWidth;
  DOUBLE centreHeight; // height of centre for polynomial zooming
  DOUBLE zoomExponent;
  DOUBLE invZoomExponent;
  DOUBLE invElevationZoomExponent;
  DOUBLE minScaledElev;
  DOUBLE invScaledElevRange;
  stringstream definition;
  bool showProgress;

  // some helper functions for the parser
  void skip(CommentedInput& text, const char* s) const;
  void skip(CommentedInput& text, const char c) const;
  string generatorName(int g) const;
  typedef std::vector< DOUBLE > Equation;
  static void writeAreaEquation(ostream& s, const Equation& e);
  static void writeAreaEquations(ostream& s, const std::vector< Equation >& ee);

    public:
  void parse(CommentedInput& text);
  bool isContinuous() const;
  DOUBLE diameterLowerBound() const;  

  // calculate coordinates of point at start of subsegment indexed by string:
  ElevatedPoint operator()(std::string index) const; 

    private:
  // some helper functions for the sample point generator:
  ElevatedPoint zoom(const ElevatedPoint& p);
  void buildBridge(
    PointCollector& pp, 
    Point a, Point b, DOUBLE h,
    int orientation
  );
  void generatePoints(
    PointCollector& pp, 
    int generator,
    const Similarity& s,
    bool& jump, // must be set to true if jumping in; this includes
               // the initial call; the method will set the variable
               // to true (or false) if (not) jumping on the way out.
    int& jumpOrientation
  );

    public:
  void applySketchSettings(DOUBLE sampleLength);
  void applyUserSettings(const Options& options);
  void generatePoints(
    PointCollector& pp, // object to process the points generated
    const Similarity& s = Similarity() // transformation to apply
  );
  void report(ostream& o);
};

enum CoordinateSystem 
{
  planarCoordinates, // Cartesian
  cubicCoordinates   // 3D points representing projection on plane x+y+z = 0
};

void readSegmentParameters(
  // input:
  stringstream& text,
  // output:
  DOUBLE& x,
  DOUBLE& y,
  int& g, // generator, must be preset to number of generators minus 1,
          // and will only be read if larger than 0
  int& d,
  int& r,
  // context:
  CoordinateSystem c = planarCoordinates
);


/*********************************************************
* Options headers                                        *
*********************************************************/

struct OptionSyntax 
{
  int rank;
  string id;
  bool requiresArgument;
  bool multipleAllowed;
  OptionSyntax(
    int _rank, 
    string _id, 
    bool _requiresValue = false,
    bool _multipleAllowed = false
  );
};

class View
{
    public:
  std::vector< Similarity > orientation;
  bool ridge;
  DOUBLE ridgeLine;
  bool defaultCentre;
  DOUBLE centreHeight;
  DOUBLE altitude;
  DOUBLE azimuth;
  DOUBLE distance;

  View();
  void reset();
  void parse(CommentedInput& text, const IFS& ifs);
  void report(ostream& s) const;

    private:
  static const OptionSyntax syntax[];
};

struct Options
{
    public:
  string ifsFileName; string ifsName; IFS ifs; 
  string cfgFileName; string cfgName; View view;
  DOUBLE accuracy;
  BridgeStyle bridgeStyle;
  bool widePassages;
  DOUBLE bridgeWidth; 
  DOUBLE bridgeCoreWidth;
  DOUBLE clearance;
  DOUBLE strength;
  DOUBLE foundation; // where to put the base of the model
  int nrColourClasses;
  int erosionRadius;
  DOUBLE resolution;
  DOUBLE parapet;
  bool supported; // supported by rock, as opposed to floating
  DOUBLE zoomExponent;
  bool printable;
  static int verbosity;

  Options();
  void reset();
  void parse(const int argc, const char* argv[]);
  void report(ostream& s) const;
  static void usage();

    private:
  static const OptionSyntax syntax[];
  static void cutId(const string id, string& fileName, string& name);
};


/*********************************************************
* Hex Grid headers (model creation)                      *
*********************************************************/

// in rendering the curve, four coordinate systems are involved,
// with coinciding origins:
// - sample space (x,y,h) (equal to model space, see below,
//   up to a scaling factor such that it fits on the grid)
// - grid space (row,col,h), where row and col are in 
//   [-gir, +gir]. (gir = gridIndexRadius)
// - model space (x,y,h): (c = cornerDepth)
//   grid (-(gir-1),-(gir-1),h) -> model ( 0,  sqrt(3) c, h)
//   grid (-(gir-1),+(gir-1),h) -> model (-c,  0,         h)
//   grid (+(gir-1),+(gir-1),h) -> model ( 0, -sqrt(3) c, h)
//   grid (+(gir-1),-(gir-1),h) -> model ( c,  0,         h)
// - output space: model (x,y,h) -> output (-y,h,-x)

// neighbours of hex cells are numbered 0 to 5, in 
// counterclockwise order, starting with the one in 
// direction (+1,0) in grid space, which is direction 
// (1, -sqrt(3)) in sample/model space

// array indexed modulo n
template < class T, int n > class WrapArray
{
    private:
  T a[n];
    public:
  WrapArray();
  T& operator[](int i);
  const T& operator[](int i) const;
};

// grids
const int gridIndexDifferenceToNeighbour[6][2] = {
  {1,0}, {1,-1}, {0,-1}, {-1,0}, {-1,1}, {0,1}
};

template < class T > 
class Grid
{
    private:
  int minr, maxr, minc, maxc, width;
  T* grid;

    public:
  class iterator
  {
      private:
    Grid< T >* host;
    int r, c, minr, maxr, minc, maxc;
    string description;
    T* current;
    iterator(
      Grid< T >* _host, 
      int _minr, int _maxr, int _minc, int _maxc,
      string _description
    );
      public:
    // check if operator is past the end (return true if not)
    bool operator()() const;

    // advance to next grid cell:
    iterator& operator++();

    // obtain reference to current grid cell:
    T& operator*() const;

    // obtain coordinates of current grid cell
    // (location().x = row; location().y = column):
    Point location() const;

    // obtain reference to neighbour number i
    // (when numbered in counterclockwise order, starting from 
    // the neighbour in the next row, in the same column):
    T& hexNeighbour(int i);
 
    friend class Grid< T >;
  };
  Grid();
  void grow(int _minr, int _maxr, int _minc, int _maxc);

  // obtain reference to cell in specified row and column:
  T& operator()(int r, int c);
  const T& operator()(int r, int c) const;

  // obtain iterator (see above) to iterate through the whole grid
  // (if a non-empty description is given, it is used for a 
  // progress indicator):
  iterator begin(string description = string());

  // obtain iterator (see above) to iterate through the whole grid
  // except for the cells on the boundary:
  iterator visitInterior(string description = string());

  // obtain width of the grid (number of rows - 1 or number of 
  // columns - 1, whatever is larger):
  int edgeLength() const;

  // obtain (row, column) coordinates of corner of the grid
  // (0: front; 1: left; 2: back; 3: right)
  Point corner(int i) const;

  // obtain (row, column) coordinates of centre of the grid
  // (could be fractional if #rows or #columns is even)
  Point centre() const;

  ~Grid();
  friend class iterator;
};

struct Sample
// An elevation sample
{
  DOUBLE h;
  SampleCategory a;
  std::pair< DOUBLE, DOUBLE > range;
  Sample();
  Sample(AnnotatedPoint p); 
  Sample(DOUBLE _h, SampleCategory _a = sampleFill);
  DOUBLE weight() const;   // size of range
  DOUBLE priority() const; // weight of jump core samples, 0 otherwise 
  bool operator<(const Sample& s) const; // sorts top-down
  bool operator==(const Sample& s) const; 
};

typedef std::vector< Sample > SampleList;

struct Slab;

typedef std::vector< Slab > SlabList;

struct Slab
// Models a solid object that covers an entire grid cell between
// specified bottom and top elevations
{
  DOUBLE bottom, top;
  Slab();
  Slab(DOUBLE _bottom, DOUBLE _top);

  // decide what walls (from top to bottom) and what parapet
  // (i.e. its height) needs to be created for this slab on the
  // boundary with the specified neighbouring grid cell:
  void decideWallAndParapet(
    const SlabList& neighbourSlabs, 
    std::vector< DOUBLE >& wallStartStop, 
    DOUBLE& parapet
  ) const;
};

class PointCollector
{
    protected:
  AnnotatedPoint lastPoint;
    public:
  virtual PointCollector& operator<<(const AnnotatedPoint& p);
  virtual bool empty() const;
  AnnotatedPoint back() const;
};

class PointRecorder: public PointCollector
{
    private:
  std::vector< Point >& pp;
    public:
  PointRecorder(std::vector< Point >& _pp);
  virtual PointCollector& operator<<(const AnnotatedPoint& p);
  virtual bool empty() const;
};

class SampleCollector: public PointCollector
{
    private:
  int minr, maxr, minc, maxc;
  int cellsPerUnit;
  // add samples with the following function in order of non-decreasing elevation
  virtual void addSampleToCell(int r, int c, const Sample& s);
  virtual void calculateSlabs(int r, int c, SlabList& slabs);
  void registerAccess(int r, int c);

    public:
  SampleCollector(int _cellsPerUnit);
  virtual PointCollector& operator<<(const AnnotatedPoint& p);
  virtual bool empty() const;
  void calculateSlabs(Grid< SlabList >& slabs, int padding);
};

class MultilevelCollector: public SampleCollector
{
    private:
  struct CellCollector
  {
    SlabList slabs;
    DOUBLE priority;
  };
  Grid< CellCollector > grid;
  DOUBLE strength, clearance, foundation;
  virtual void addSampleToCell(int r, int c, const Sample& s);
  virtual void calculateSlabs(int r, int c, SlabList& slabs);

    public:
  MultilevelCollector(
    int _cellsPerUnit, DOUBLE _strength, DOUBLE _clearance, DOUBLE _foundation
  );
};

class MeanCollector: public SampleCollector
{
    private:
  struct CellCollector 
  {
    DOUBLE sum;
    DOUBLE weight;
    int count;
    CellCollector();
  };
  Grid< CellCollector > grid;
  DOUBLE foundation;
  virtual void addSampleToCell(int r, int c, const Sample& s);
  virtual void calculateSlabs(int r, int c, SlabList& slabs);

    public:
  MeanCollector(int _cellsPerUnit, DOUBLE _foundation);
};

// add slabs to grid cells to construct the upper reference plane:
void constructUpperPlane(
  Grid< SlabList >& slabs, DOUBLE foundation, int discreteRidge
);

// calculate camera position *before* rotation according to azimuth 
// and altitude settings (rotation is done in the COLLADA output, so
// that the viewing direction rotates with it):
ElevatedPoint cameraPosition(
  Grid< SlabList >& slabs,
  const Similarity& ModelCoordFromGridCoord,
  const View& view
);

// erosion operator:
void smoothen(Grid< SlabList >& slabs, DOUBLE foundation);

#ifdef NARROWS
void widenPassages(Grid< SlabList >& slabs, 
  DOUBLE cellSpacing, DOUBLE strength, DOUBLE clearance, DOUBLE factor
);
#endif
#ifdef PRINTABLE
void makePrintable(Grid< SlabList >& slabs,
  DOUBLE cellSpacing, DOUBLE strength, DOUBLE clearance
);
#endif

class MeshVertex;

class Mesh
// A class to generate COLLADA geometries (meshes) of a specified
// material. To add a polygon, add vertices with addVertex, then
// call endPolygon(). Meshes are output and cleared when clear()
// is called, or when the destructor is called. Blender refuses to
// load meshes with huge coordinate arrays, so before adding a 
// polygon, use full() to check whether the mesh is getting too 
// large (and if so, call clear()). (This approach is sufficient 
// as long as polygons are small). The material names for all meshes
// that are output are collected in an array which can be accessed
// with materials(); the COLLADA nodes library should create a node
// for each mesh and bind the material "SOLID" to the material name
// in materials(); (the main function does this.)
{
    private:
  string material;
  std::vector< ElevatedPoint > coordinates;
  std::vector< int > polygonSizes;
  std::vector< int > polygons;
  int nrPending; // number of vertices in unfinished polygon
  static std::vector< string > theMaterials;
    public:
  Mesh(const string& _material = "");
  void setMaterial(const string& _material);
  static const std::vector< string >& materials();

  // for each vertex, a normal index should be specified:
  // 0-5: pointing to neighbours in directions 0-5
  // 6:   pointing upwards
  // 7:   pointing downwards 
  void addVertex(MeshVertex& v, int normalIndex);
  void addVertex(const Point& p, DOUBLE h, int normalIndex);
  void addVertex(DOUBLE x, DOUBLE y, DOUBLE h, int normalIndex);

  void endPolygon();
  bool full() const;
  void clear();
  ~Mesh();
};

class MeshVertex
// A class for re-usable vertices in a mesh. The class keeps 
// track of an index in the coordinates array of the mesh. Only 
// the first time the vertex is used, the coordinates are added 
// to the coordinates array. Note: MeshVertex objects become 
// invalid if passed to different Mesh objects, or whenever the 
// Mesh is cleared.
{
    private:
  ElevatedPoint point;
  int index;
    public:
  MeshVertex(Point p = Point(), DOUBLE h = 0);
  friend class Mesh;
};

// create geometries (meshes) for the COLLADA geometry library:
void drawModel(
  Grid< SlabList >& slabs,
  const Similarity& ModelCoordFromGridCoord,
  bool printable, 
  int nrColours, 
  DOUBLE parapetHeight = 0
);
void drawLowerReferencePlane();
void drawUpperReferencePlane(
  Grid< SlabList >& slabs,
  const Similarity& ModelCoordFromGridCoord,
  int discreteRidge
);


/*********************************************************
* main header                                            *
*********************************************************/

int main(const int argc, const char* argv[]);


/*********************************************************
**********************************************************
**                                                      **
** IMPLEMENTATION                                       **
**                                                      **
**********************************************************
*********************************************************/


/*********************************************************
* Input/Output                                           *
*********************************************************/

const char* Complaint::errorLabel[5] = 
{ 
  "WARNING", 
  "Error",
  "Error while parsing IFS", 
  "Error", 
  "Internal error" 
};

Complaint::Complaint(ErrorLevel _errorLevel): 
  errorLevel(_errorLevel) 
{
};
  
Complaint::Complaint(
  ErrorLevel _errorLevel, 
  const string& expected, 
  const string& read
):
  errorLevel(_errorLevel)
{
  complaint << "expected \"" << expected << "\", but read \"" << read << "\".";
}

Complaint::~Complaint()
{
  cerr << endl << errorLabel[(int) errorLevel] << ": " << complaint.str() << endl;
  if (errorLevel == usageError) 
    cerr << "Run pftrail without arguments for usage." << endl;
  if (errorLevel > warning) exit((int) errorLevel); 
}

template <typename T> Complaint& Complaint::operator<<(const T& t)
{
  complaint << t; return *this;
}

void CommentedInput::fillBuffer()
{
  string text;
  while (text.empty())
  {
    std::getline(source, text);
    if (source.fail()) return;
    // remove long comments
    if (text.substr(0,2) == "/*")
    {
      text.erase(0,2);
      while (text.find("*/") > text.length())
      {
        std::getline(source, text);
        if (source.fail()) 
        {
          Complaint(warning) << "unclosed comment in file";
          return;
        }
      }
      text.erase(0, text.find("*/") + 2);
    }
    // remove short comments:
    text.erase(std::min(text.find("//"), text.length()));
  }
  buffer.clear();
  buffer.str(text);
}

CommentedInput::CommentedInput(string& fileName):
  source(fileName.c_str())
{};

bool CommentedInput::fail() const
{
  return buffer.fail() || source.fail();
}

void CommentedInput::clear()
{
  buffer.clear();
}

template < class T > CommentedInput& CommentedInput::operator>>(T& t)
{
  if (source.fail()) return *this;
  buffer >> t;
  while (buffer.fail() && buffer.eof() && !source.fail()) 
  {
    fillBuffer();
    buffer >> t;
  }
  return *this;
}

void CommentedInput::getLine(string& s)
{
  if (source.fail()) return;
  std::getline(buffer >> std::ws, s);
  if (s.empty()) 
  {
    fillBuffer();
    std::getline(buffer >> std::ws, s);
  }
}

void CommentedInput::getLowerCaseWord(string& s)
{
  *this >> s; makeLowerCase(s);
}

bool CommentedInput::find(const char* keyword, const string& name)
{
  string lowerCaseKeyword(keyword); makeLowerCase(lowerCaseKeyword);
  string lowerCaseName(name); makeLowerCase(lowerCaseName);
  string inputWord;
  while (!fail() && inputWord != lowerCaseName)
  {
    while (!fail() && inputWord != lowerCaseKeyword) 
      getLowerCaseWord(inputWord);
    if (fail()) return false;
    getLowerCaseWord(inputWord);
  }
  return (!fail() && inputWord == lowerCaseName);
}

void makeLowerCase(string& s)
{
  for (int i = 0; i < s.length(); ++i) s[i] = tolower(s[i]);
}

DOUBLE strtofstop(const string& str)
{
  stringstream s(str);
  DOUBLE x;
  s >> x >> std::ws;
  if (!s.eof()) Complaint(userError)
    << "expected a floating-point number but read \"" << str << "\".";
  return x; 
}

int strtoistop(const string& str)
{
  stringstream s(str);
  int n;
  s >> n >> std::ws;
  if (!s.eof()) Complaint(userError)
    << "expected an integer but read \"" << str << "\".";
  return n; 
}

void reportProgress(DOUBLE x, string description)
{
  if (Options::verbosity < 1 || description.empty()) return;
  if (x >= 1)
    cerr << description << ": done." << string(progressUnits - 3, ' ') << endl;
  else
  {
    cerr << description << ": [";
    for (int i = 1; i < progressUnits; ++i)
    {
      if (x * progressUnits >= i)
        cerr << (i % 5 ? '-' : '+');
      else
        cerr << (i % 5 ? ' ' : ':');
    }
    cerr << "]\r";
  }
  cerr.flush();
}


/*********************************************************
* Geometry                                               *
*********************************************************/

Point::Point(DOUBLE _x, DOUBLE _y): 
  x(_x), y(_y) 
{};

Point Point::operator+(const Point& p) const 
{
  return Point(x + p.x, y + p.y);
}

Point& Point::operator+=(const Point& p) 
{
  x += p.x; y += p.y; return *this;
}

Point Point::operator-(const Point& p) const 
{
  return Point(x - p.x, y - p.y);
}

Point& Point::operator-=(const Point& p) 
{
  x -= p.x; y -= p.y; return *this;
}

Point Point::operator-() const 
{
  return Point(-x, -y);
}
  
DOUBLE Point::operator*(const Point& p) const 
{
  return x * p.x + y * p.y;
}

Point Point::operator*(const DOUBLE d) const 
{
  return Point(x * d, y * d);
}
 
Point& Point::operator*=(const DOUBLE d) 
{
  x *= d; y *= d; return *this;
}

Point Point::operator/(const DOUBLE d) const 
{
  return Point(x / d, y / d);
}
 
Point& Point::operator/=(const DOUBLE d) 
{
  x /= d; y /= d; return *this;
}

DOUBLE Point::sqLength() const 
{
  return x * x + y * y;
}

DOUBLE Point::length() const 
{
  return sqrt(sqLength());
}

ostream& operator<<(ostream& s, const Point& p)
{
  return s << '(' << p.x << ',' << p.y << ')';
}

ElevatedPoint::ElevatedPoint(): 
  Point(), h(0)  
{};

ElevatedPoint::ElevatedPoint(const Point& _p, DOUBLE _h): 
  Point(_p), h(_h)
{};

ElevatedPoint::ElevatedPoint(DOUBLE _x, DOUBLE _y, DOUBLE _h):
  Point(_x, _y), h(_h)
{};

ElevatedPoint ElevatedPoint::operator+(const ElevatedPoint& e) const 
{
  return ElevatedPoint(x + e.x, y + e.y, h + e.h);
}

ElevatedPoint& ElevatedPoint::operator+=(const Point& p) 
{
  x += p.x; y += p.y; return *this;
}

ElevatedPoint& ElevatedPoint::operator+=(const ElevatedPoint& e) 
{
  x += e.x; y += e.y; h += e.h; return *this;
}

ElevatedPoint ElevatedPoint::operator-(const ElevatedPoint& e) const 
{
  return ElevatedPoint(x - e.x, y - e.y, h - e.h);
}

ElevatedPoint& ElevatedPoint::operator-=(const Point& p) 
{
  x -= p.x; y -= p.y; return *this;
}

ElevatedPoint& ElevatedPoint::operator-=(const ElevatedPoint& e) 
{
  x -= e.x; y -= e.y; h -= e.h; return *this;
}

ElevatedPoint ElevatedPoint::operator*(const DOUBLE d) const 
{
  return ElevatedPoint(x * d, y * d, h * d);
}
 
ElevatedPoint ElevatedPoint::operator/(const DOUBLE d) const 
{
  return ElevatedPoint(x / d, y / d, h / d);
}
 
ostream& operator<<(ostream& s, const ElevatedPoint& p)
{
  return s << '(' << p.x << ',' << p.y << ',' << p.h << ')';
}

AnnotatedPoint::AnnotatedPoint():
  ElevatedPoint(), a(sampleFill), range(-1,-1)
{};

AnnotatedPoint::AnnotatedPoint(
  const ElevatedPoint& _e, SampleCategory _a, DOUBLE r
):
  ElevatedPoint(_e), a(_a), range(_e.h - r, _e.h + r)
{};

AnnotatedPoint::AnnotatedPoint(
  const Point& _p, DOUBLE _h, SampleCategory _a, DOUBLE r
):
  ElevatedPoint(_p, _h), a(_a), range(h - r, h + r)
{};

// Below are helper functions for the computation of smallest 
// enclosing circles, computing the centre and squared radius 
// of a minimum enclosing circle for a set of points with 
// 1, 2 or 3 points on the boundary given.

// three boundary points given:
void computeMinDisk(
  Point p1, Point p2, Point p3, 
  Point& centre, DOUBLE& sqRadius
)
{
  // check if a pair determines the diameter:
  for (int i = 0; i < 3; ++i)
  {
    centre = (p1+p2)/2;
    sqRadius = (p2-p1).sqLength()/4 + practicallyZero;
    if ((p3-centre).sqLength() <= sqRadius) return;
    std::swap(p1,p2); std::swap(p2,p3);
  }
  // centre of bisectors satisfies:
  // (p2-p1)*2*centre = ||p2||^2 - ||p1||^2
  DOUBLE c[3][3]; // coefficients of the three equations
  c[0][0] = (p2.x-p1.x)*2; c[0][1] = (p2.y-p1.y)*2; c[0][2] = p2.sqLength()-p1.sqLength(); 
  c[1][0] = (p3.x-p2.x)*2; c[1][1] = (p3.y-p2.y)*2; c[1][2] = p3.sqLength()-p2.sqLength(); 
  c[2][0] = (p1.x-p3.x)*2; c[2][1] = (p1.y-p3.y)*2; c[2][2] = p1.sqLength()-p3.sqLength(); 
  // solve:
  if (abs(c[2][0]) > abs(c[1][0])) std::swap(c[1], c[2]);
  if (abs(c[1][0]) > abs(c[0][0])) std::swap(c[0], c[1]);
  c[0][1] /= c[0][0]; c[0][2] /= c[0][0]; c[0][0] = 1;
  for (int i = 1; i < 3; ++i)
    for (int j = 1; j < 3; ++j)
      c[i][j] -= c[i][0] * c[0][j];
  // 1st column is henceforth to be ignored, should be regarded as 1,0,0
  if (abs(c[2][1]) > abs(c[1][1])) std::swap(c[1], c[2]);
  c[1][2] /= c[1][1];
  c[0][2] -= c[0][1] * c[1][2];
  // 2nd column is henceforth to be ignored, should be regarded as 0,1,0
  centre = Point(c[0][2], c[1][2]);
  sqRadius = (p1-centre).sqLength();
}

// two boundary points given:
void computeMinDisk(std::vector< Point >& pp, 
  const Point& p1, const Point& p2,
  Point& centre, DOUBLE& sqRadius
)
{
  computeMinDisk(pp[0], p1, p2, centre, sqRadius);
  for (int i = 1; i < pp.size(); ++i)
    if ((pp[i]-centre).sqLength() > sqRadius) 
      computeMinDisk(p1, p2, pp[i], centre, sqRadius);
}

// one boundary point given:
void computeMinDisk(std::vector< Point >& pp, 
  const Point& p1,
  Point& centre, DOUBLE& sqRadius
)
{
  computeMinDisk(pp[0], pp[1], p1, centre, sqRadius);
  for (int i = 2; i < pp.size(); ++i)
    if ((pp[i]-centre).sqLength() > sqRadius) 
      computeMinDisk(pp, p1, pp[i], centre, sqRadius);
}

DOUBLE radius(std::vector< Point >& points)
{
  if (points.empty()) return 0;
  std::vector< Point >& pp = points;
  std::vector< Point > threepoints;

  // ensure minimum input size of three points:
  if (pp.size() < 3)
  {
    threepoints = points;
    while (threepoints.size() < 3) threepoints.push_back(threepoints[0]); 
    pp = threepoints;
  }

  // permute randomly:
  for (int i = 0; i < pp.size(); ++i) 
    std::swap(pp[i], pp[i + (rand() % (pp.size()-i))]);

  // solve using incremental construction:
  Point centre; DOUBLE sqRadius;
  computeMinDisk(pp[0], pp[1], pp[2], centre, sqRadius);
  for (int i = 3; i < pp.size(); ++i)
    if ((pp[i]-centre).sqLength() > sqRadius) 
      computeMinDisk(pp, pp[i], centre, sqRadius);
  return sqrt(sqRadius);
}

Rotation::Rotation(const Point& _x, const Point& _y): 
  x(_x), y(_y) 
{};

Rotation Rotation::transpose() const 
{
  return Rotation(Point(x.x, y.x), Point(x.y, y.y));
}

Rotation Rotation::operator*(const Rotation& r) const 
{
  Rotation t = r.transpose();
  return Rotation(
    Point( x * t.x, x * t.y ),
    Point( y * t.x, y * t.y ) 
  );
}

Rotation Rotation::operator*=(const Rotation& r) 
{
  Rotation t = r.transpose();
  x = Point( x * t.x, x * t.y );
  y = Point( y * t.x, y * t.y );
  return *this;
}

Point Rotation::operator*(const Point& p) const 
{
  return Point(x * p, y * p);
}

Point Rotation::operator()(const Point& p) const
{
  return operator*(p);
}

DOUBLE Rotation::sqScale() const
{
  return x.x * x.x + y.x * y.x;
}

DOUBLE Rotation::scale() const
{
  return sqrt(sqScale()); 
}

int Rotation::orientation() const
{
  DOUBLE o = x.x * y.y;
  if (o < 0) return -1;
  if (o > 0) return  1;
  return 0;
}

Similarity::Similarity(
  const Rotation& _rotation, 
  const ElevatedPoint& _translation, 
  DOUBLE _scale
):
  rotation(_rotation), translation(_translation), scale(_scale) 
{};

Similarity Similarity::operator*(const Similarity& s) const 
{
  return Similarity(
    rotation * s.rotation,
    operator()(s.translation),
    scale * s.scale
  );
}

Similarity& Similarity::operator*=(const Similarity& s) 
{
  translation = operator()(s.translation);
  rotation *= s.rotation;
  scale *= s.scale;
  return *this;
}

Similarity& Similarity::apply(const Similarity& s) 
{
  translation = s(translation);
  rotation = s.rotation * rotation;
  scale *= s.scale;
  return *this;
}

Point Similarity::operator()(const Point& p) const 
{
  return rotation * p + (Point) translation;
}

ElevatedPoint Similarity::operator()(const ElevatedPoint& e) const 
{
  return ElevatedPoint(rotation * (Point) e, scale * e.h) + translation;
}

void Similarity::print(
  std::ostream& o, const string& leftLabel, const string& rightLabel
) const
{
  o << leftLabel << std::fixed << std::setprecision(3)
    << "| " << std::setw(6) << rotation.x.x << ' '
            << std::setw(6) << rotation.x.y << ' '
            << "  0   " << ' '
            << std::setw(6) << translation.x << " |" << rightLabel << endl;
  o << std::setw(leftLabel.length() + 2)
    << "| " << std::setw(6) << rotation.y.x << ' '
            << std::setw(6) << rotation.y.y << ' '
            << "  0   " << ' '
            << std::setw(6) << translation.y << " |" << endl;
  o << std::setw(leftLabel.length() + 2) 
    << "| " << "  0   " << ' '
            << "  0   " << ' '
            << std::setw(6) << scale << ' '
            << std::setw(6) << translation.h << " |" << endl;
  o << std::setw(leftLabel.length() + 2) 
    << "| " << "  0   " << ' '
            << "  0   " << ' '
            << "  0   " << ' '
            << "  1   " << " |" << endl;
  o.unsetf(std::ios_base::floatfield);
}

template < class T > 
const Rotation TrianglePoint< T >::TriToCar = 
  Rotation(Point(1.0, 0.5), Point(0, sqrt((DOUBLE) 0.75)));

template < class T >
TrianglePoint< T >::TrianglePoint(int s, T a, T b) 
{
  x[s%3] = a; x[(s+1)%3] = b; x[(s+2)%3] = -(a+b);
}

template < class T >
TrianglePoint< T >::TrianglePoint(const Point& p) 
{
  x[0] =  p.x - SQRTTHIRD * p.y;
  x[1] = -p.x - SQRTTHIRD * p.y;
  x[2] = -(x[0] + x[1]); 
};

template < class T >
T TrianglePoint< T >::operator[](int i) const 
{ 
  return x[i % 3]; 
}

template < class T >
TrianglePoint<T>::operator Point() const 
{
  return TriToCar * Point(x[0], x[2]);
}

TrianglePoint< int > round(const TrianglePoint< DOUBLE >& t) 
{
  DOUBLE roundingError[3];
  for (int i = 0; i < 3; ++i) roundingError[i] = round(t[i]) - t[i];

  // first find coordinate with smallest difference to rounded number
  int closest = 0;
  for (int i = 1; i < 3; ++i) {
    if (abs(roundingError[i]) >= abs(roundingError[closest])) continue;
    closest = i;
  }
  // round this coordinate; distribute the rounding error over the other two
  // and round them accordingly

  TrianglePoint< int > r(closest,
    (int) round(t[closest]),
    (int) round(t[closest+1] - 0.5*roundingError[closest])
  );

  return r;
}


/*********************************************************
* IFS (parsing, analysis and sampling)                   *
*********************************************************/

Segment::Segment(Similarity _transformation, int _generator):
  transformation(_transformation),
  generator(_generator)
{};

void IFS::skip(CommentedInput& text, const char* s) const
{
  std::string aword;
  text.getLowerCaseWord(aword);
  if (text.fail() || aword != s) Complaint(ifsParsingError, s, aword);
}

void IFS::skip(CommentedInput& text, const char c) const
{
  char acharacter;
  text >> acharacter; acharacter = tolower(acharacter);
  if (text.fail() || acharacter != c) 
    Complaint(ifsParsingError, string() + c, string() + acharacter);
}

string IFS::generatorName(int g) const
{
  if (g == 0)  return string("principal generator");
  else         return string("generator ") + (char) ('A' + g);
}

void IFS::writeAreaEquation(ostream& s, const Equation& e)
{
  bool zero = true;
  for (int i = 0; i < e.size()-1; ++i)
  {
    if (abs(e[i]) < practicallyZero) continue;
    if (zero)
    {
      if (e[i] < 0) s << '-';
    }
    else s << (e[i] < 0 ? " - " : " + ");
    zero = false;
    DOUBLE coefficient = abs(e[i]);
    if (coefficient != 1) s << coefficient << ' '; 
    s << "f(" << (char) ('A' + i) << ')';
  }
  if (zero) s << '0';
  s << " = " << std::setprecision(10) << e.back();
}

void IFS::writeAreaEquations(ostream& s, const std::vector< Equation >& ee)
{
  for (std::vector< Equation >::const_iterator e = ee.begin(); e != ee.end(); ++e)
  {
    writeAreaEquation(s, *e); 
    s << endl;
  }
}

void IFS::parse(CommentedInput& text)
{
  definition.clear();
  std::string aword;

  // skip any (alternative) names of the IFS
  while (!text.fail()) 
  {
    text.getLowerCaseWord(aword);
    if (aword == "ifs") text >> aword; else break;
  }

  // determine coordinate system
  Rotation grid;
  CoordinateSystem coordinateSystem = planarCoordinates;
  if (aword == "square")
    grid = Rotation(Point(1, 0), Point(0, 1));
  else if (aword == "triangular")
    grid = Rotation(Point(1, 0.5), Point(0, SQRTTHREE * 0.5));
  else if (aword == "cubic" || aword == "barycentric")
  {
    grid = Rotation(Point(1.5, 0), Point(-SQRTTHREE * 0.5, -SQRTTHREE));
    coordinateSystem = cubicCoordinates;
  }    
  else 
    Complaint(ifsParsingError, "square\", \"triangular\", or \"cubic", aword);
  skip(text, "grid");

  // the definition is recorded in Cartesian coordinates, 
  // regardless of the input format
  definition << "square grid" << endl;

  // determine number of generators
  int n; text >> n;
  if (text.fail()) Complaint(ifsParsingError) 
    << "could not read number of generators or segments.";
  int nrGenerators;
  bool multipleGeneratorSyntax;
  text.getLowerCaseWord(aword);
  if (aword == "segments") {
    multipleGeneratorSyntax = false;
    nrGenerators = 1;
    definition << n << ' ' << "segments" << endl;
  }
  else if (aword == "generators") {
    multipleGeneratorSyntax = true;
    nrGenerators = n;
    definition << n << ' ' << "generators" << endl;
  }
  else
    Complaint(ifsParsingError, "segments\" or \"generators", aword);
  
  // prepare for collecting area equations and continuity status 
  continuous = true;
  std::vector< Equation > areaEquations;
  if (Options::verbosity >= 2)
  {
    cerr << "f(x):= area filled divided by squared distance between endpoints of generator x." << endl
         << "The following equations are derived from the generators:" << endl;
  }

  // read generators one by one
  for (int i = 0; i < nrGenerators; ++i)
  {
    bool areaGiven = false;
    DOUBLE givenArea;
    if (multipleGeneratorSyntax)
    {
      skip(text, "generator");
      char generatorLabel = (char) ('a' + i);
      skip(text, generatorLabel); 
      skip(text, ':');
      definition << endl << "generator " << generatorLabel << ':' << endl;

      text.getLowerCaseWord(aword);
      if (aword == "fills")
      {
        text >> givenArea;
        if (text.fail()) Complaint(ifsParsingError) 
          << "could not read area filled by " << generatorName(i) << ".";
        if (givenArea < 0) Complaint(ifsParsingError)
          << "negative area specified for " << generatorName(i) << ".";
        areaGiven = true; 
        definition << "fills " << givenArea << endl;
        text.getLowerCaseWord(aword);
      }

      n = strtoistop(aword);
      if (text.fail() || n == 0) Complaint(ifsParsingError)
        << "could not read number of segments of " << generatorName(i) << ".";
      skip(text, "segments");
      definition << n << " segments" << endl;
    }

    if (n < 2) Complaint(ifsParsingError)
      << "less than two segments specified for " << generatorName(i) << ".";
    skip(text, "segment"); skip(text, "values:");
    definition << endl << "segment values:" << endl;

    // initialize the generator
    generators.push_back(Generator());
    Generator& generator = generators.back();
    Point location(0,0);
    areaEquations.push_back(Equation(nrGenerators+1));

    // read the segments one by one
    for (int j = 0; j < n; )
    {
      int segmentLabel; text >> segmentLabel;
      if (text.fail()) 
      {
        text.clear(); char c; text >> c; 
        Complaint(ifsParsingError)
          << "expected segment index " << j << ", "
          << "but read \"" << c << "\".";
      }
      ++j;
      if (segmentLabel != j) Complaint(ifsParsingError) 
        << "expected segment index " << j << ", "
        << "but read " << segmentLabel << ".";

      // read segment parameters
      int g = nrGenerators - 1;
      DOUBLE x, y; int d, r;
      skip(text, ':');
      string segmentParameters;
      text.getLine(segmentParameters);
      stringstream segmentParameterStream(segmentParameters);
      readSegmentParameters(segmentParameterStream, x, y, g, d, r, coordinateSystem);
      string aword;
      segmentParameterStream >> aword;
      if (!aword.empty()) 
        Complaint(userError) << "unexpected text \"" << aword << "\" on segment line.";
      Point edge = grid * Point(x, y);
      if (abs(edge.x) < practicallyZero) edge.x = 0;
      if (abs(edge.y) < practicallyZero) edge.y = 0;
      
      // record what was read (paraphrased)
      definition << j << ":  " << edge.x << ", " << edge.y << ", ";
      if (nrGenerators > 1) definition << (g == -1 ? 'Y' : (char) ('A'+ g)) << ", ";
      definition << d << ", " << r << endl;

      // construct segment or jump
      ElevatedPoint anchor(location, 0);
      location += edge;
      if (g == -1)
      {
        continuous = false;
        generator.segmentIndex.push_back(-1);
      }
      else
      {
        if (d == -1) 
        { 
          edge = -edge; 
          anchor = ElevatedPoint(location, 0); 
          r = -r; 
        }
        // for the vertical length of plane-filling segments, 
        // only the sign (d) matters at this point; 
        // the magnitude will be determined later:
        generator.segmentIndex.push_back(generator.segments.size());
        generator.segments.push_back(
          Segment(
            Similarity(
              Rotation(
                Point(edge.x, -r * edge.y), 
                Point(edge.y,  r * edge.x)
              ),
              anchor,
              d
            ),
            g
          )
        );
        areaEquations.back()[g] -= edge.sqLength();
      }
      generator.orientation.push_back(r);
    }

    // verify that the generator has non-zero length
    DOUBLE generatorSize = location.sqLength();
    if (generatorSize == 0) Complaint(userError) 
      << "distance between end points of " << generatorName(i) << " is zero"; 

    // finish the area equations
    areaEquations.back()[i] += location.sqLength();
    if (Options::verbosity >= 2)
    { 
      cerr << "- from " << generatorName(i) << ": ";
      writeAreaEquation(cerr, areaEquations.back());
      cerr << endl;
    }
    if (areaGiven) 
    {
      areaEquations.push_back(Equation(nrGenerators+1));
      areaEquations.back()[i] = generatorSize;
      areaEquations.back()[nrGenerators] = givenArea;
      if (Options::verbosity >= 2) 
      {
        cerr << "- specified area filled by " << generatorName(i) << ": ";
        writeAreaEquation(cerr, areaEquations.back());
        cerr << endl;
      }
    }

    // rotate and scale such that endpoints are (0,0) and (1,0)
    Similarity normalization(
      Rotation(
        Point( location.x/generatorSize, location.y/generatorSize),
        Point(-location.y/generatorSize, location.x/generatorSize)
      ),
      ElevatedPoint(0, 0, 0),
      (DOUBLE) 1.0
    );
    for (
      std::vector< Segment >::iterator s = generator.segments.begin(); 
      s != generator.segments.end(); 
      ++s
    )
      s->transformation.apply(normalization);
  }

  if (areaEquations.size() == nrGenerators) 
  {
    // no area was explicitly specified, w.l.o.g. assume f(A) = 1
    areaEquations.push_back(Equation(nrGenerators+1));
    areaEquations.back()[0] = 1;
    areaEquations.back()[nrGenerators] = 1;
    if (Options::verbosity >= 2)
    { 
      cerr << "- for a suitable choice of area unit we may assume: ";
      writeAreaEquation(cerr, areaEquations.back());
      cerr << endl;
    }
  }

  // solve the area equations; implementation is suboptimal but "n" is really small
  for (int j = nrGenerators-1; j >= 0; --j)
  {
    // find an equation i with coefficients j+1...n-1 all zero, 
    // and among those, the one with largest absolute value for coefficient j
    int bestEquation = -1; DOUBLE divisor = 0; 
    for (int i = 0; i < areaEquations.size(); ++i)
    {
      if (abs(areaEquations[i][j]) <= abs(divisor)) continue;
      bool eligible = true; 
      for (int k = j+1; eligible && k < nrGenerators; ++k)
        eligible = (areaEquations[i][k] == 0);
      if (!eligible) continue;
      bestEquation = i;
      divisor = areaEquations[i][j]; 
    }
    if (abs(divisor) <= practicallyZero)
    {
      if (Options::verbosity > 0) 
      {
        cerr << "Area equations reduce to:" << endl;
        writeAreaEquations(cerr, areaEquations);
      }
      Complaint(userError) << "cannot determine area filled by " << generatorName(j) << ".";
    }
    if (bestEquation != j) swap(areaEquations[bestEquation], areaEquations[j]);

    // normalize:
    for (int k = 0; k <= nrGenerators; ++k) areaEquations[j][k] /= divisor;

    // eliminate generator j from all other rows
    for (int i = 0; i < areaEquations.size(); ++i)
    { 
      if (i == j) continue;
      DOUBLE multiplier = areaEquations[i][j];
      for (int k = 0; k <= nrGenerators; ++k) 
        areaEquations[i][k] -= multiplier * areaEquations[j][k];
    }
  }
  // now, the coefficients matrix should be in the form
  // |   1   0   ...   0   f(A)   |
  // |   0   1   ...   0   f(B)   |
  // |   :   :    :    :    :     |
  // |   0   0   ...   1   f(N)   |
  // |   0   0   ...   0    0     |
  // |   :   :    :    :    :     |

  // check for inconsistencies:
  for (int i = nrGenerators; i < areaEquations.size(); ++i)
    if (abs(areaEquations[i].back()) > practicallyZero) 
    {
      if (Options::verbosity > 0) 
      {
        cerr << "Area equations reduce to:" << endl;
        writeAreaEquations(cerr, areaEquations);
      }
      Complaint(userError) << " specified areas are inconsistent.";
    }
  
  for (int i = 0; i < nrGenerators; ++i)
    if (abs(areaEquations[i].back()) <= practicallyZero) 
    {
      if (Options::verbosity > 0) 
      {
        cerr << "Area equations reduce to:" << endl;
        writeAreaEquations(cerr, areaEquations);
      }
      Complaint(userError) 
        << generatorName(i) << " does not generate a plane-filling curve.";
    }

  if (Options::verbosity >= 2 && nrGenerators >= 2)
  {
    cerr << "(f(A), ..., f(" << (char) ('A' + nrGenerators-1) << ")) = (";
    cerr << std::setprecision(3) << areaEquations[0].back();
    for (int i = 1; i < nrGenerators; ++i)
      cerr << ", " << std::setprecision(3) << areaEquations[i].back();
    cerr << ')' << endl;
  }

  // fill in the third rows of the transformation matrices
  for (int i = 0; i < nrGenerators; ++i)
  {
    DOUBLE t = 0;
    for (
      std::vector< Segment >::iterator s = generators[i].segments.begin(); 
      s != generators[i].segments.end(); 
      ++s
    )
    {
      s->transformation.scale *= 
        s->transformation.rotation.sqScale() * 
        areaEquations[s->generator].back() / areaEquations[i].back();
      if (s->transformation.scale > 0)
      {
        s->transformation.translation.h += t;
        t += s->transformation.scale;
      }
      else // s->transformation.scale <= 0
      {
        t -= s->transformation.scale;
        s->transformation.translation.h += t;
      }
    }
  }

  if (Options::verbosity >= 2) 
  {
    cerr << "The following transformation matrices have been constructed:" << endl;
    for (int i = 0; i < nrGenerators; ++i) {
      if (nrGenerators > 1) cerr << "Generator " << (char) ('A' + i) << ':' << endl;
      const Generator& g = generators[i];
      for (int s = 0; s < g.segmentIndex.size(); ++s) 
      {
        if (g.segmentIndex[s] == -1) { 
          cerr << "+ (";
          switch(g.orientation[s])
          {
            case -1: cerr << "left-aligned "; break;
            case  1: cerr << "right-aligned "; break;
            default: break;
          }
          cerr << "jump)" << endl; 
          continue; 
        }
        const Segment& segment = g.segments[g.segmentIndex[s]];
        string rightLabel = " " + generatorName(segment.generator);
        segment.transformation.print(cerr, "+ ", rightLabel);
      }
    }
  }

  // calculate lower bound on diameter and, for each generator, upper bound 
  // on "radius" (max. distance from curve point to closest endpoint) 
  // (these numbers are needed to calculate the sampling density) 
  if (Options::verbosity >= 2)
    cerr << "Generating sketches for all generators to estimate shape parameters..." << endl;
  // apply settings for rough sketches  
  const DOUBLE maxEdgeLength = 0.05;
  applySketchSettings(maxEdgeLength); 

  DOUBLE overallMaxRadius = 0;
  for (int i = 0; i < nrGenerators; ++i) {
    Generator& g = generators[i];
    std::vector< Point > pp;
    PointCollector* collector = new PointRecorder(pp);
    bool skipJump = true;
    int jumpOrientation = 0;
    generatePoints(*collector, i, Similarity(), skipJump, jumpOrientation);
    if (i == 0) diameter = radius(pp) * 2.0;
    // find maximum squared distance from any sample point to closest end point
    g.maxRadius = 0; 
    for (std::vector< Point >::const_iterator p = pp.begin(); p != pp.end(); ++p) 
      g.maxRadius = std::max(g.maxRadius,
        std::min(
          (*p - Point(0,0)).sqLength(), 
          (*p - Point(1,0)).sqLength())
      );
    // calculate actual distance
    g.maxRadius = sqrt(g.maxRadius);
    overallMaxRadius = std::max(overallMaxRadius, g.maxRadius);
  }
  // calculate maximum further expansion in recursion
  DOUBLE expansion = overallMaxRadius * maxEdgeLength / (1.0 - maxEdgeLength);
  for (int i = 0; i < nrGenerators; ++i)
    generators[i].maxRadius += expansion;

  if (Options::verbosity >= 2)
  {
    cerr << "... done sketching. The following numbers were derived:" << endl;
    cerr << " * the diameter divided by the distance between endpoints is at least " << diameter << ";" << endl;
    cerr << " * for any point on the curve";
    if (nrGenerators > 1) cerr << "s generated by generators A..." << (char) ('A' + nrGenerators-1);
    cerr << ", the distance to the closest endpoint divided by the distance between endpoints is at most";
    if (nrGenerators == 1)
      cerr << ' ' << generators[0].maxRadius << '.' << endl;
    else 
    { 
      cerr << ", respectively" << endl << generators[0].maxRadius;
      for (int i = 1; i < nrGenerators; ++i) cerr << ", " << generators[i].maxRadius;
      cerr << '.' << endl;
    }
  }
}

bool IFS::isContinuous() const
{
  return continuous;
}

DOUBLE IFS::diameterLowerBound() const
{
  return diameter;
}

ElevatedPoint IFS::operator()(std::string index) const
{
  if (index[0] - '0' == generators[0].segments.size()) return ElevatedPoint(Point(1,0), 1);
  Similarity s;
  int whichEnd = 0;
  int g = 0;
  for (int i = 0; i < index.length() && index[i] != ' ' && index[i] != '\t'; ++i) 
  {
    const Generator& generator = generators[g]; 
    if (s.scale == 0) Complaint(userError) 
      << "IFS::operator() called on \"" << index.substr(0,i) << "...\", but \""
      << index.substr(0,i) << "\" denotes curve section with zero two-dimensional content.";
    int j = index[i] - '0';
    if (j < 0 || j >= generator.segmentIndex.size()) Complaint(userError) 
      << "IFS::operator() called on \"" << index.substr(0,i+1) << "...\", but \""
      << index[i] << "\" does not denote a subsegment of \"" << index.substr(0,i) << "\".";
    if (s.scale < 0) j = generator.segmentIndex.size() - 1 - j;
    if (generator.segmentIndex[j] == -1) Complaint(userError) 
      << "IFS::operator() called on \"" << index.substr(0,i) << "...\", but \""
      << index.substr(0,i) << "\" denotes a jump.";
    const Segment& segment = generator.segments[generator.segmentIndex[j]];
    if (segment.transformation.scale < 0) whichEnd = 1 - whichEnd;
    s *= segment.transformation;
    g = segment.generator;
  }
  return s(ElevatedPoint(Point(whichEnd,0), whichEnd));
}

ElevatedPoint IFS::zoom(const ElevatedPoint& p)
{
  Point pz; DOUBLE hz;

  // scale distance x to x^{1/a}
  DOUBLE oldDistance = p.length();
  DOUBLE newDistance = pow(oldDistance, 1.0/zoomExponent);
  if (newDistance == 0) pz = Point(0,0);
  else pz = ((Point) p) * newDistance / oldDistance;

  // same for elevation, but with additional scaling (to stay within 0-1 range)
  // and with exponent 1/(2a-1) instead of exponent a (much stronger zoom),
  // to compensate for the fact that small parts of the curve are "flatter"
  DOUBLE oldVerticalDistance = abs(p.h);
  DOUBLE newVerticalDistance = pow(oldVerticalDistance, invElevationZoomExponent);
  if (newVerticalDistance == 0) hz = 0;
  else hz = invScaledElevRange * 
    ((p.h < 0 ? -newVerticalDistance : newVerticalDistance) - minScaledElev);

  return ElevatedPoint(pz, hz);
}
 
void IFS::buildBridge(
  PointCollector& pp, 
  Point a, Point b, const DOUBLE h,
  int orientation
)
{
  // calculate length and width in steps of size (roughly) cell diameter * sqrt(3)/4 
  Point d = b - a;
  DOUBLE length = d.length();
  int nrSamples = (int) (length / bridgeSampleDistance) + 1;
  DOUBLE halfwidth = bridgeWidthFactor * pow(d.sqLength(), 1.0/6.0);
  int corew = (bridgeCoreWidth == 0 ? -1 : 
    bridgeCoreWidth / (bridgeSampleDistance * 2.0));
  
  // if asymmetric, move all width to one side:
  int leftSide = (orientation > -1 ? 1 : 0);
  int rightSide = (orientation < 1 ? 1 : 0);
  if (leftSide + rightSide == 1) { halfwidth *= 2.0; corew *= 2; }

  // calculate single step size:
  Point lngStep = d / nrSamples;
  Point latStep = Point(lngStep.y, -lngStep.x); // side step 

  for (int i = 0; i <= nrSamples; ++i)
  {
    // calculate local width w (steps left/right of centre line)
    int w = -1;
    switch((int) bridgeStyle) 
    { 
      case pointedBridges: 
        w = (int) (
          pow(pow((DOUBLE) nrSamples, 4.0) - pow((DOUBLE) 2*i-nrSamples, 4.0), 0.25) 
          * halfwidth / nrSamples + 0.5
        );
        break;
      case rectangularBridges:
        w = (int) (halfwidth + 0.5);
        break;
      default:
        Complaint(internalError) << "unrecognized connection type or bridge style";
    }
    Point c = a; c -= (latStep * (leftSide * w));
    for (int j = -leftSide * w; j <= rightSide * w; ++j)
    {
      pp << AnnotatedPoint(
        c, h, 
        ( -corew <= j && j <= corew ? sampleJumpCore : sampleJump ), 
        length * length
      ); 
      c += latStep;
    }
    a += lngStep;
  }
}

void IFS::generatePoints(
  PointCollector& pp, 
  int generator, 
  const Similarity& s,
  bool& jump,
  int& jumpOrientation
) 
{
  ElevatedPoint p[2];
  p[s.scale >= 0 ? 0 : 1] = s(ElevatedPoint(Point(0,0), 0));
  p[s.scale >= 0 ? 1 : 0] = s(ElevatedPoint(Point(1,0), 1));
  Generator& g = generators[generator];
  bool mustRefine = false;

  if (zoomExponent == 0)
    mustRefine = s.rotation.sqScale() > g.maxSqSampleLength;
  else 
  {
    // Points at distance x are moved to x^{1/z}. This appears to lead to the following:
    // Let X be the origin (centre of zoom), A be any point, and r a fixed radius.
    // Points at distance r from A that are furthest from A after zooming are found:
    // -  if r <= 2 ||XA||: on the circle around X with radius ||XA||;
    //    The distance after zooming is:  r ||XA||^{1/z-1}
    // -  if r >= 2 ||XA||: on the ray from A through X.
    //    The distance after zooming is:  ||XA||^{1/z} + (r-||XA||)^{1/z}
    // The effect of zooming is stronger near the origin, so what matters is the case:
    // A is the segment endpoint closest to X; r is the segment's maxRadius.
 
    Point A = p[0]; DOUBLE sqDistA = A.sqLength();
    Point B = p[1]; DOUBLE sqDistB = B.sqLength();
    if (sqDistA < sqDistB) { A = B; sqDistA = sqDistB; }
    // A is now the segment endpoint furthest from X. 
    DOUBLE distA = sqrt(sqDistA);

    DOUBLE r = g.maxRadius * s.rotation.scale();
    if (r <= distA * 2.0) 
    {
      // first a test that is sufficient if distA is really small, to avoid 
      // division by zero or overflow in the real test
      if (pow(distA, invZoomExponent) <= curveSampleRadius)
        mustRefine = false;
      else 
        mustRefine = (r * pow(distA, invZoomExponent-1.0) > curveSampleRadius);
    }
    else 
      mustRefine = 
        (pow(distA, invZoomExponent) + pow(r - distA, invZoomExponent) > curveSampleRadius);
  }
  if (!mustRefine)   
  {
    if (zoomExponent != 0) p[1] = zoom(p[1]);
    if (jump)
    {
      if (zoomExponent != 0) p[0] = zoom(p[0]);
      if (bridgeStyle != noBridges && !pp.empty() && 
        // hack that prevents tunnels from one rendering to the next 
        pp.back().h - p[0].h <= 0.1 
      ) 
        buildBridge(pp, pp.back(), p[0], p[0].h, jumpOrientation);
      pp << AnnotatedPoint(p[0], sampleFill);
    }
    pp << AnnotatedPoint(p[1], sampleFill);
    jump = false;
    return;
  }
  int starti, endi, stepi;
  if (s.scale > 0) { starti = 0; endi = g.segmentIndex.size(); stepi = 1; }
  else             { starti = g.segmentIndex.size()-1; endi = -1; stepi = -1; }
  for (int i = starti; i != endi; i += stepi)
  {
    if (g.segmentIndex[i] == -1) 
    {
      // jump orientation is only applied to jumps that are segments in the IFS;
      // if consecutive jumps are shortcut, the orientation is set to zero:
      jumpOrientation = (jump ? 0 : stepi * s.rotation.orientation() * g.orientation[i]);
      jump = true;
      continue;
    }
    const Segment& e = g.segments[g.segmentIndex[i]];
    generatePoints(pp, e.generator, s * e.transformation, jump, jumpOrientation);
  }
  if (p[1].h - p[0].h > 0.01 && showProgress) 
    reportProgress(min((DOUBLE) 0.999, p[1].h), "Sampling the traversal");
}

void IFS::applySketchSettings(DOUBLE sampleLength)
{
  centreHeight = 0;
  zoomExponent = 0;
  invZoomExponent = 1;
  invElevationZoomExponent = 1;
  curveSampleRadius = sampleLength; // should not have any effect
  for (std::vector< Generator >::iterator g = generators.begin(); g != generators.end(); ++g)
  {
    g->maxSampleLength = sampleLength;
    g->maxSqSampleLength = pow(g->maxSampleLength, 2.0);
  }
  bridgeStyle = noBridges;
  showProgress = false;
}

void IFS::applyUserSettings(const Options& options)
{
  zoomExponent = options.zoomExponent;
  centreHeight = options.view.centreHeight;
  invZoomExponent = (DOUBLE) 1.0 / max((DOUBLE) 1.0, zoomExponent);
  invElevationZoomExponent = (DOUBLE) 1.0 / elevationZoomExponent(zoomExponent);
  minScaledElev = -pow(centreHeight, invElevationZoomExponent);
  DOUBLE maxScaledElev =  pow((DOUBLE) 1.0-centreHeight, invElevationZoomExponent);
  invScaledElevRange = (DOUBLE) 1.0 / (maxScaledElev - minScaledElev);
  curveSampleRadius = 0.25 * pow(diameter, invZoomExponent) / options.resolution / options.accuracy;
  for (std::vector< Generator >::iterator g = generators.begin(); g != generators.end(); ++g)
  {
    g->maxSampleLength = curveSampleRadius / g->maxRadius;
    g->maxSqSampleLength = pow(g->maxSampleLength, 2.0);
  }
  bridgeStyle = options.bridgeStyle;
  bridgeSampleDistance = SQRTTHREE * 0.25 * diameter / options.resolution;
  bridgeWidthFactor = 0.5 * options.bridgeWidth * pow(diameter, 1.0/3.0) / bridgeSampleDistance;
  bridgeCoreWidth = options.bridgeCoreWidth;
  showProgress = true;
}

void IFS::generatePoints(PointCollector& pp, const Similarity& s)
{
  bool jump = true;
  int jumpOrientation = 0;

  // when zooming in, in an attempt to make the precision of DOUBLE last longest
  // (especially w.r.t. elevation), we shift the traversal so that the centre of 
  // zoom lies in the origin;
  Similarity shifted = s; 
  if (zoomExponent != 0) shifted.translation.h -= centreHeight;

  if (showProgress) reportProgress(0, "Sampling the traversal");
  generatePoints(pp, 0, shifted, jump, jumpOrientation);
  // the zoom function that is used by generatePoints undoes the shift so that the 
  // output is where it should be
  if (showProgress) reportProgress(1, "Sampling the traversal");
}

void IFS::report(ostream& o)
{
  o << "IFS definition:" << endl << endl << definition.str() << endl;
}

void skipChar(istream& text, const char c)
// only used by readSegmentParameters
{
  char acharacter;
  text >> acharacter; acharacter = tolower(acharacter);
  if (text.fail() || acharacter != c) 
    Complaint(ifsParsingError, string() + c, string() + acharacter);
}

void readSegmentCoordinates(
  stringstream& text,
  DOUBLE& x, 
  DOUBLE& y, 
  CoordinateSystem c
)
{
  text >> x;
  if (text.fail()) Complaint(ifsParsingError) << "could not read segment's x-coordinate.";

  skipChar(text, ',');
  text >> y;
  if (text.fail()) Complaint(ifsParsingError) << "could not read segment's y-coordinate.";

  if (c == cubicCoordinates)
  {
    DOUBLE z;
    skipChar(text, ',');
    text >> z;
    if (text.fail()) Complaint(ifsParsingError) << "could not read segment's z-coordinate.";
    z += x; z += y; z /= 3.0;
    x -= z; y -= z;
  }
}

void readSegmentParameters(
  stringstream& text,
  DOUBLE& x,
  DOUBLE& y,
  int& g, // generator, must be preset to number of generators minus 1,
          // and will only be read if larger than 0
          // is set to -1 if the segment is a bridge
  int& d,
  int& r,
  CoordinateSystem c
)
{
  readSegmentCoordinates(text, x, y, c);

  bool readGenerator = (g > 0);
  if (readGenerator)
  {
    skipChar(text, ',');
    char gchar;
    text >> gchar;
    if (text.fail()) Complaint(ifsParsingError) << "could not read segment's generator.";
    if (gchar == 'X') { 
      g = -1; d = 0; r = 0; return;
    }
    else if (gchar == 'Y') 
      g = -1;
    else if (gchar < 'A' || gchar > 'A' + g) 
      Complaint(ifsParsingError) << "invalid generator identifier \"" << gchar << "\".";
    else g = gchar - 'A';
  }

  skipChar(text, ',');
  text >> d;
  if (text.fail()) Complaint(ifsParsingError) << "could not read segment's direction indicator.";
  if ((d*d > 1) || (readGenerator && g >= 0 && d*d != 1) || (g == -1 && d != 0)) 
    Complaint(ifsParsingError) << "invalid direction indicator (" << d << ").";
  if (d == 0) g = -1;

  skipChar(text, ',');
  text >> r;
  if (text.fail()) Complaint(ifsParsingError) << "could not read segment's reflection indicator.";
  if ((g >= 0 && r*r != 1) || (g == -1 && r*r > 1)) 
    Complaint(ifsParsingError) << "invalid reflection indicator (" << r << ").";
}


/*********************************************************
* Options                                                *
*********************************************************/

// first declare some helpers:

const OptionSyntax* findOption(const OptionSyntax options[], string argv);

struct Argument 
// class to store options with arguments;
// different options must have unique ranks and ids
{
  int position; // position in the input   
  const OptionSyntax* key;
  string value;
  Argument(int _position, const OptionSyntax& _key, const string& _value);

  // functions to extract a number from the string value, and 
  // verify that the number is within given bounds:
  int intValue(const int min, const int max) const;
  DOUBLE floatValue(const DOUBLE min, const DOUBLE max) const;

  // comparison operator to sort by rank, and for options of equal rank,
  // by position in the input:
  bool operator<(const Argument& a) const;
};


// and here is the implementation:

OptionSyntax::OptionSyntax(
  int _rank, 
  string _id, 
  bool _requiresValue,
  bool _multipleAllowed
):
  rank(_rank), 
  id(_id), 
  requiresArgument(_requiresValue),
  multipleAllowed(_multipleAllowed)
{
};

View::View() 
{
  reset();
}

void View::reset()
{
  orientation.clear();
  orientation.push_back(Similarity());
  ridge = true;
  ridgeLine = 0.5;
  centreHeight = 0;
  defaultCentre = true;
  altitude = 18.5;
  azimuth = 0;
  distance = 1.1;
}

void View::parse(CommentedInput& text, const IFS& ifs)
{
  reset();
  ElevatedPoint centre;

  // collect view/rendering instructions
  std::vector< Argument > arguments;
  string inputWord;
  while (true)
  {
    text >> inputWord;
    if (text.fail()) break;
    const OptionSyntax* o = findOption(syntax, inputWord);
    if (!o) 
      Complaint(userError) << "unrecognized keyword \"" << inputWord << "\" in rendering/view configuration.";
    if (o->id == "ifs") break;
    if (o->id == "view" && !arguments.empty()) break;
    if (o->requiresArgument)
    {     
      text.getLine(inputWord);
      if (inputWord.empty()) 
        Complaint(userError) << "missing value for \"" << o->id << "\" in rendering/view configuration.";
    }
    if (o->id != "view") arguments.push_back(
      Argument(arguments.size(), *o, inputWord)
    );           
  }

  // sort and check for doubles
  std::sort(arguments.begin(), arguments.end());
  for (int i = 2; i < arguments.size(); ++i)
    if (
      arguments[i].key->rank == arguments[i-1].key->rank && 
      arguments[i].value != arguments[i-1].value &&
      !arguments[i].key->multipleAllowed
    )
      Complaint(userError) << 
      "multiple values for \"" << arguments[i].key->id + "\" (\"" << 
      arguments[i-1].value << "\" and \"" <<
      arguments[i].value << "\").";

  // process values
  orientation.clear();
  DOUBLE unitLength = 1.0;
  for (std::vector< Argument >::const_iterator a = arguments.begin(); a != arguments.end(); ++a)
  {
    if (a->key->id == "render") 
    {
      DOUBLE x, y; int g, d, r; g = 0;
      stringstream value(a->value);
      Point origin;
      if (value.peek() == '@')
      {
        skipChar(value, '@');
        value >> origin.x;
        if (value.fail()) Complaint(userError) 
          << "could not read x-coordinate of starting point on \"render @\" line";
        skipChar(value, ',');
        value >> origin.y;
        if (value.fail()) Complaint(userError) 
          << "could not read y-coordinate of starting point on \"render @\" line";
        skipChar(value, ',');
      }
      readSegmentParameters(value, x, y, g, d, r);
      if (g == -1) Complaint(userError) 
        << "generating line segment on \"render\" line must have non-zero orientation.";
      inputWord.clear(); value >> inputWord;
      if (!inputWord.empty()) Complaint(userError) 
        << "unexpected text \"" << inputWord << "\" on \"render\" line.";
      DOUBLE l = Point(x,y).length();
      if (l == 0) Complaint(userError) << "generating line segment on \"render\" line must have non-zero length.";
      if (orientation.empty()) unitLength = l;
      x /= unitLength; y /= unitLength; origin /= unitLength;
      orientation.push_back(Similarity(
        Rotation(
          Point(d * x, -r * y),
          Point(d * y,  r * x)
        ),
        d == 1 ? ElevatedPoint(origin              , 0)
               : ElevatedPoint(origin + Point(x, y), 1),       
        d
      ));
      if (orientation.size() == 1)
        ridgeLine = origin.x + 0.5 * x; // note: cannot undo ridge command, because cmnds are sorted, render first.
    }
    else if (a->key->id == "ridge")    ridgeLine = a->floatValue(-1,1000);
    else if (a->key->id == "centre")
    {
      defaultCentre = false;
      centre = ifs(a->value);
    }
    else if (a->key->id == "azimuth")  azimuth = a->floatValue(-360,360);
    else if (a->key->id == "altitude") altitude = a->floatValue(-90,90);
    else if (a->key->id == "distance") distance = a->floatValue(-2,2);
  }
  if (orientation.empty()) orientation.push_back(Similarity());
  centre = orientation[0](centre);
  // modify orientation to move centre to origin w.r.t. x- and y-coordinates
  for (
    std::vector< Similarity >::iterator o = orientation.begin();
    o != orientation.end();
    ++o
  ) 
    o->translation -= (Point) centre; // conversion to point drops h-coordinate

  // record height coordinate of centre
  centreHeight = centre.h;

  if (ridgeLine < 0) ridge = false; else ridgeLine -= centre.x;
}

void View::report(ostream& s) const 
{
  s.unsetf(std::ios_base::floatfield);
  s << "View settings:" << endl;
  orientation[0].print(s, "transformation       "); 
  for (
    std::vector< Similarity >::const_iterator o = ++orientation.begin(); 
    o != orientation.end(); 
    ++o
  )
    o->print(s, "and                  ");
  s.unsetf(std::ios_base::floatfield);
  s << "upper ref. plane     ";
  if (ridge)         s << "starts at line x = " << ridgeLine << endl;
  else               s << "none" << endl;
  if (defaultCentre) s << "camera centre        automatic" << endl;
  else               s << "camera/zoom centre   (0,0," << centreHeight << ')' << endl;
  s << "camera azimuth       " << azimuth << endl;
  s << "camera altitude      " << altitude << endl;
  s << "camera distance      " << distance << " (1 = where it sees all)" << endl; 
  s << endl;
}

const OptionSyntax View::syntax[9] =
{
  OptionSyntax(0, "render", true, true),
  OptionSyntax(1, "ridge", true),
  OptionSyntax(2, "centre", true),
  OptionSyntax(3, "altitude", true),
  OptionSyntax(4, "azimuth", true),
  OptionSyntax(5, "distance", true),
  OptionSyntax(6, "ifs", false),
  OptionSyntax(7, "view", true),
  OptionSyntax(99, "")
};

int Options::verbosity = 1;
 
Options::Options()
{
  reset();
}

void Options::reset() 
{
  view.reset();
  accuracy = 1;
  bridgeStyle = pointedBridges;
  widePassages = false;
  bridgeWidth = 0.015; 
  bridgeCoreWidth = 0.010;
  clearance = 0.010;
  strength = 0.005;
  foundation = floorHeight - strength;
  nrColourClasses = 1;
  erosionRadius = -1;
  resolution = 200;
  parapet = 0.002;
  verbosity = 1;
  supported = true;
  printable = false;
  zoomExponent = 0; // 0 means 1, but with more efficient non-zooming algo
}

void Options::parse(const int argc, const char* argv[])
{
  // restore defaults
  reset();

  // collect command line arguments, check syntax
  std::vector< Argument > arguments;
  for (int i = 1; i < argc; ++i)
  {
    if (argv[i][0] != '-') 
    {
      arguments.push_back(Argument(arguments.size(), syntax[0], argv[i]));
      continue;
    }
    const OptionSyntax* o = findOption(syntax, string(argv[i]));
    if (!o) Complaint(usageError) << "unrecognized option \"" << argv[i] << "\".";
    if (o->requiresArgument)
    {     
      ++i; 
      if (i == argc) Complaint(usageError) << "missing value for \"" << o->id << "\" option.";
    }
    arguments.push_back(Argument(arguments.size(), *o, argv[i]));           
  }

  // sort command line arguments, check for missing or double arguments
  std::sort(arguments.begin(), arguments.end());
  if (arguments.size() < 1 || arguments[0].key->rank != 0) 
    Complaint(usageError) << "missing input file name.";
  if (arguments.size() > 1 && arguments[1].key->rank == 0)
    Complaint(usageError) << 
      "multiple inputs specified (\"" << 
      arguments[0].value << "\" and \"" <<
      arguments[1].value << "\").";
  for (int i = 2; i < arguments.size(); ++i)
    if (
      arguments[i].key->rank == arguments[i-1].key->rank && 
      arguments[i].value != arguments[i-1].value &&
      !arguments[i].key->multipleAllowed
    )
      Complaint(usageError) << 
      "multiple values for \"" << arguments[i].key->id + "\" option (\"" << 
      arguments[i-1].value << "\" and \"" <<
      arguments[i].value << "\").";
 
  // process command line arguments
  for (std::vector< Argument >::const_iterator a = arguments.begin(); a != arguments.end(); ++a)
  {
    switch(a->key->id[1]) 
    {
      case 'i': cutId(a->value, ifsFileName, ifsName); break;
      case 's':
#ifdef PRINTABLE
        if (a->value == "3d")
        {
          printable = true;
          supported = true;
          widePassages = true;
          parapet = 0;
        }
        else 
#endif
        if (a->value == "eroded")
        { 
          erosionRadius = 0;
          supported = true;
          parapet = 0;
        }
        else if (a->value == "floating") 
          supported = false;
        else if (a->value != "solid")
          Complaint(usageError) << "unrecognized style \"" << a->value << "\".";
        break;
      case 'r': resolution = a->floatValue(5, 10000); break;
      case 'a': accuracy = a->floatValue(0.25, 20.0); break;
      case 'b': bridgeWidth = a->floatValue(0, 0.1); break;
      case 'c': cutId(a->value, cfgFileName, cfgName); break;
      case 'f': bridgeStyle = rectangularBridges; break;
      case 'g': nrColourClasses = a->intValue(1, 1000); break;
      case 'h': clearance = a->floatValue(0, 1); break;
      case 'k': 
        if (erosionRadius < 0) Complaint(usageError) << 
          "\"-k\" option invalid for \"solid\" or \"floating\" style.";
        erosionRadius = a->intValue(0, resolution);
        break;
      case 'm': bridgeCoreWidth = a->floatValue(0, 0.1); break;
#ifdef NARROWS
      case 'n': widePassages = true; break;
#endif
      case 'p':
        if (erosionRadius >= 0) Complaint(usageError) <<
          "\"-p\" option invalid for \"eroded\" style.";
        if (printable) Complaint(usageError) <<
          "\"-p\" option invalid for \"3d\" stytle.";
        parapet = a->floatValue(0, 1);
        break;
      case 'q': verbosity = 0; break;
      case 'v': 
        if (verbosity != 1) Complaint(usageError) << "conflicting options \"-q\" and \"-v\"."; 
        verbosity = 2;
        break;
      case 'w': 
        strength = a->floatValue(0, 1); 
        foundation = floorHeight - strength; 
        break;
      case 'z': zoomExponent = a->floatValue(1, 10); break;
      default: Complaint(internalError) << "unrecognized option " << a->key->id << ".";
    } 
  }

  // read IFS and view configuration
  CommentedInput ifsFile(ifsFileName);
  if (ifsFile.fail()) Complaint(userError) << "cannot read file \"" << ifsFileName << "\".";
  if (!ifsName.empty() && !ifsFile.find("IFS", ifsName))
    Complaint(userError) << "IFS \"" << ifsName << "\" not found in file \"" << ifsFileName << "\".";
  if (verbosity >= 2) 
  {
    cerr << "Reading IFS";
    if (!ifsName.empty()) cerr << " \"" << ifsName << "\"";
    cerr << " from file \"" << ifsFileName << "\"..." << endl;
  } 
  ifs.parse(ifsFile);
  cerr << "IFS read successfully." << endl;
  view.parse(ifsFile, ifs);

  if (!cfgFileName.empty() || !cfgName.empty()) 
  {
    if (cfgFileName.empty()) cfgFileName = ifsFileName;
    CommentedInput cfgFile(cfgFileName);
    if (cfgFile.fail()) Complaint(userError) << "cannot read file \"" << cfgFileName << "\".";
    if (!cfgName.empty() && !cfgFile.find("view", cfgName))
      Complaint(userError) << "View \"" << cfgName << "\" not found in file \"" << cfgFileName << "\".";
    if (verbosity >= 2) 
    {
      cerr << "Reading view";
      if (!cfgName.empty()) cerr << " \"" << cfgName << "\"";
      cerr << " from file \"" << cfgFileName << "\"...";
      cerr.flush();
    } 
    view.parse(cfgFile, ifs);
    if (verbosity >= 2) cerr << " done." << endl;
  }

  if (parapet > clearance * 0.5) 
    Complaint(warning) << "Parapet height is set to more than half of the minimum vertical distance between layers (minimum tunnel height). Beware of unexpected artefacts.";

  if (view.ridge && !supported) 
  { 
    Complaint(warning) 
      << "no upper reference plane and ridge are drawn with \"floating\" style.";
    view.ridge = false;
  }
  if (view.ridge && printable) 
  { 
    Complaint(warning) 
      << "no reference planes and ridge are drawn with \"3d\" style.";
    view.ridge = false;
  }
 
  if (zoomExponent != 0) 
  {
    if (view.defaultCentre)
      Complaint(userError) << "rendering/view configuration does not specify centre for polynomial zoom."; 
    if (view.ridge) 
    {
      Complaint(warning) << "no upper reference plane and ridge are drawn in polynomial zoom.";
      view.ridge = false;
    }
    if (zoomExponent > 1 && !ifs.isContinuous())
      Complaint(warning) << "jumps are drawn as straight line segments despite polynomial zoom.";
  }

  ifs.applyUserSettings(*this);

  if (verbosity >= 2) 
  {
    cerr << endl;
    report(cerr);
    view.report(cerr);
  }
}

void Options::report(ostream& s) const
{
  s.unsetf(std::ios_base::floatfield);
  s << "IFS file/name        " << ifsFileName; 
  if (!ifsName.empty()) s << ':' << ifsName;
  s << endl;
  s << "View file/name       ";
  if (cfgFileName.empty()) s << "IFS default";
  else { s << cfgFileName; if (!cfgName.empty()) s << ':' << cfgName; }
  s << endl;

  s << endl;
  s << "General settings:" << endl;
  s << "zoom exponent        "; 
  if (zoomExponent == 0) s << "no zoom"; else s << zoomExponent; 
  s << endl;

  s << "style                ";
  if (printable) s << "3d";
  else if (!supported) s << "floating";
  else if (erosionRadius >= 0) s << "eroded";
  else s << "solid";
  s << endl; 
 
  s << "resolution           " << (int) resolution      << endl;
  s << "accuracy             " << accuracy        << endl;
  s << "number of colours    " << nrColourClasses << endl;

  if (erosionRadius >= 0) 
    s << "erosion radius       " << erosionRadius   << endl;
      
  if (erosionRadius < 0) 
  {
    if (!printable) s << "parapet height       " 
        << std::setprecision(4) << parapet         << endl;
    s << "min. tunnel height   " 
      << std::setprecision(4) << clearance       << endl;
    s << "min. thickness       " 
      << std::setprecision(4) << strength        << endl;
  }
  s << "bridge/tunnel style  ";
  if (ifs.isContinuous())
    s << "not applicable (traversal is continuous)" << endl;
  else
  {
    switch (bridgeStyle) 
    { 
      case noBridges: s << "none"; break;
      case pointedBridges: s << "pointed"; break;
      case rectangularBridges: s << "rectangular"; break;
      default: Complaint(internalError) << "bridge style undefined";
    }
    s << endl;
  }
#ifdef NARROWS
  if (!ifs.isContinuous() || widePassages)
    s << "brdg/tnl/passg width " << bridgeWidth     << endl;
#else
  if (!ifs.isContinuous()) s << "bridge/tunnel width  " << bridgeWidth     << endl;
#endif
  if (!ifs.isContinuous())
    s << "brdg/tnl core width  " << bridgeCoreWidth << endl;
#ifdef NARROWS
  s << "widened narrows      " << (widePassages ? "yes" : "no") << endl;
#endif

  s << endl;
  s.unsetf(std::ios_base::floatfield);
} 

void Options::usage() 
{
  cerr << pftrailInfo << endl
       << "Usage: pftrail <ifsfile>[:<ifsid>] [<options>]" << endl
       << "Output: geometries and nodes libraries. Concatenate preamble, output, and postamble"
       << " to produce a COLLADA file that can be imported in Blender." << endl
       << endl
       << "<ifsfile>   name of a file with an iterated function system in extended" << endl
       << "            Ventrella format (see documentation)" << endl
       << "<ifsid>     name of the IFS, if the file contains more than one" << endl
       << "<options>   the following options are supported:" << endl
       << "   -a <number>   accuracy, use values under 1 for undersampling (default: 1)" << endl
       << "   -b <number>   maximum bridge/tunnel width (default: 0.015)" << endl
       << "   -c [<ifsfile>][:<viewid>]   name of view configuration in ifs-file" << endl
       << "                 at least the file name or the view id must be specified" << endl
       << "                 (default: the configuration that follows the curve definition)" << endl
       << "   -f            use fixed-width instead of pointed bridges" << endl
       << "   -g <number>   number of colour classes in colour gradient (default: 0)" << endl
       << "   -h <number>   minimum tunnel height (default: 0.010)" << endl
       << "   -k <number>   size of the kernel in eroded style (default: 0)" << endl
       << "   -m <number>   width of bridge/tunnel core/median strip (default: 0.010)" << endl
#ifdef NARROWS
       << "   -n            widen narrow passages if defined in IFS (default: do not)" << endl
#endif
       << "   -r <number>   resolution: image diameter / grid cell diameter (default: 200)" << endl
       << "   -p <number>   height of parapets in solid/floating style (default: 0.002)" << endl
       << "   -q            quiet mode: less info on error output" << endl
       << "   -s <style>    \"solid\", "
#ifdef PRINTABLE
<< "\"3d\", "
#endif
       << "\"floating\", or \"eroded\" (default: solid; see manual)" << endl
       << "   -v            verbose mode: more info on error output" << endl
       << "   -w <number>   minimum weight (thickness) of surfaces (default: 0.005)" << endl
       << "   -z <number>   exponent for polynomial zooming (default: 1)" << endl
       << endl;
  exit(-1);
}

#ifdef NARROWS
const OptionSyntax Options::syntax[18] = 
#else
const OptionSyntax Options::syntax[17] = 
#endif
{
  OptionSyntax( 0, "-i", true),
  OptionSyntax( 1, "-s", true),
  OptionSyntax( 2, "-r", true),
  OptionSyntax( 3, "-a", true),
  OptionSyntax( 4, "-b", true),
  OptionSyntax( 5 ,"-c", true),
  OptionSyntax( 6, "-f"),
  OptionSyntax( 7, "-g", true),
  OptionSyntax( 8, "-h", true),
  OptionSyntax( 9, "-k", true),
  OptionSyntax(10, "-m", true),
#ifdef NARROWS
  OptionSyntax(11, "-n"),
#endif
  OptionSyntax(12, "-p", true),
  OptionSyntax(13, "-q"),
  OptionSyntax(14, "-v"),
  OptionSyntax(15, "-w", true),
  OptionSyntax(16, "-z", true),
  OptionSyntax(99, "") // sentinel end
}; 

void Options::cutId(const string id, string& fileName, string& name)
{
  fileName = id;
  name.erase();
  int cut = fileName.find(':');
  if (cut >= fileName.length()) return;
  name = fileName.substr(cut+1);
  fileName.erase(cut);
}

const OptionSyntax* findOption(const OptionSyntax options[], string argv)
{
  string lowerCaseArg(argv.substr(0,4)); makeLowerCase(lowerCaseArg);
  for (const OptionSyntax* o = &options[0]; ; ++o) 
  {
    if (o->id.empty()) return 0;
    if (lowerCaseArg == o->id.substr(0,4)) return o;
  }
}

Argument::Argument(int _position, const OptionSyntax& _key, const string& _value):
  position(_position), key(&_key), value(_value)
{};
 
int Argument::intValue(const int min, const int max) const
{
  int x = strtoistop(value);
  if (x < min || x > max) Complaint(usageError) 
    << "unexpected value \"" + value + "\" for \"" << key->id << "\" option.\n"
    << "Expected value between " << min << " and " << max << ".";
  return x;
}

DOUBLE Argument::floatValue(const DOUBLE min, const DOUBLE max) const
{
  DOUBLE x = strtofstop(value);
  if (x < min || x > max) Complaint(usageError) 
    << "unexpected value \"" + value + "\" for \"" << key->id << "\" option.\n"
    << "Expected value between " << min << " and " << max << ".";
  return x;
}

bool Argument::operator<(const Argument& a) const 
{ 
  if (key->rank < a.key->rank) return true;
  return (position < a.position);
}


/*********************************************************
* Hex Grid (model creation)                              *
*********************************************************/

template < class T, int n > 
WrapArray< T, n >::WrapArray() {};

template < class T, int n > 
T& WrapArray< T, n >::operator[](int i)
{
  if (i < 0) i += n * (1 - i/n);
  return a[i % n];
} 

template < class T, int n > 
const T& WrapArray< T, n >::operator[](int i) const
{
  if (i < 0) i += n * (1 - i/n);
  return a[i % n];
}

template < class T > 
Grid< T >::iterator::iterator(
  Grid< T >* _host, 
  int _minr, int _maxr, int _minc, int _maxc,
  string _description
):
  host(_host), minr(_minr), maxr(_maxr), minc(_minc), maxc(_maxc),
  r(_minr), c(_minc), 
  description(_description),
  current(&(*host)(r,c))
{
  reportProgress(0, description);
};

template < class T > 
bool Grid< T >::iterator::operator()() const
{
  return (r <= maxr);
}

template < class T > 
typename Grid< T >::iterator& Grid< T >::iterator::operator++()
{
  ++c; if (c > maxc) 
  { 
    c = minc; ++r; 
    reportProgress((DOUBLE) (r - minr)/(maxr - minr + 1), description);
  }
  if (r <= maxr) current = &(*host)(r,c);
  return *this;
}

template < class T > 
T& Grid< T >::iterator::operator*() const
{
  return *current;
}

template < class T > 
Point Grid< T >::iterator::location() const
{
  return Point(r,c);
}

template < class T > 
T& Grid< T >::iterator::hexNeighbour(int i)
{
  const int* d = gridIndexDifferenceToNeighbour[(i+6)%6];
  T& neighbour = (*host)(r + d[0], c + d[1]);
  // indexing the neighbour may have caused the grid to be resized, therefore:
  current = &(*host)(r,c);
  return neighbour;
}

template < class T > 
Grid< T >::Grid():
  minr(0), maxr(-1), minc(0), maxc(-1), width(0)
{};

template < class T > 
void Grid< T >::grow(int _minr, int _maxr, int _minc, int _maxc)
{
  if (width > 0)
  {
    _minr = min(minr, _minr);
    _maxr = max(maxr, _maxr);
    _minc = min(minc, _minc);
    _maxc = max(maxc, _maxc);
  }
  if (minr == _minr && maxr == _maxr && minc == _minc && maxc == _maxc) return;
  int _width = _maxc - _minc + 1;
/*
  if (Options::verbosity >= 2) cerr << "Growing grid to [ " 
    << _minr << ", " << _maxr << " ] x [ " << _minc << ", " << _maxc << " ]" << endl;
*/
  T* _grid = new T[(_maxr - _minr + 1) * _width];
  if (width > 0) 
  {
    for (int r = minr; r <= maxr; ++r)
      for (int c = minc; c <= maxc; ++c)
        _grid[(r - _minr) * _width + (c - _minc)] =
          grid[(r - minr) * width + (c - minc)];
    delete [] grid;
  }
  grid = _grid;
  width = _width;
  minr = _minr;
  maxr = _maxr;
  minc = _minc;
  maxc = _maxc;  
}

template < class T > 
T& Grid< T >::operator()(int r, int c)
{
  if (r < minr || r > maxr || c < minc || c > maxc) 
  {
    int _minr = minr;
    int _maxr = maxr;
    int _minc = minc;
    int _maxc = maxc;
    if (width == 0) 
    {
      _minr = r-200; _maxr = r+200; _minc = c-200; _maxc = c+200;
    }
    else
    {
      while (r < _minr) _minr = _minr - (_maxr - _minr) / 2 - 1;
      while (r > _maxr) _maxr = _maxr + (_maxr - _minr) / 2 + 1;
      while (c < _minc) _minc = _minc - (_maxc - _minc) / 2 - 1;
      while (c > _maxc) _maxc = _maxc + (_maxc - _minc) / 2 + 1;
    }
    grow(_minr, _maxr, _minc, _maxc);
  }
  return grid[(r - minr) * width + (c - minc)];
}

template < class T > 
const T& Grid< T >::operator()(int r, int c) const
{
  if (r < minr || r > maxr || c < minc || c > maxc) 
    Complaint(internalError) << "index in grid out of bounds.";
  return grid[(r - minr) * width + (c - minc)];
}

template < class T > 
typename Grid< T >::iterator Grid< T >::begin(string description)
{
  return iterator(this, minr, maxr, minc, maxc, description);
}

template < class T > 
typename Grid< T >::iterator Grid< T >::visitInterior(string description)
{
  return iterator(this, minr+1, maxr-1, minc+1, maxc-1, description);
}

template < class T > 
int Grid< T >::edgeLength() const
{
  return max(maxr - minr, maxc - minc);
}

template < class T > 
Point Grid< T >::corner(int i) const
{
  switch(i)
  {
    case 0: return Point(minr,maxc); // front corner
    case 1: return Point(minr,minc); // left corner
    case 2: return Point(maxr,minc); // back corner
    case 3: return Point(maxr,maxc); // right corner
    default: Complaint(internalError) << "invalid argument to Grid::corner(...)";
  }
  return Point();
}

template < class T >
Point Grid< T >::centre() const
{
  return Point((DOUBLE) 0.5 * (minr+maxr), (DOUBLE) 0.5 * (minc+maxc));
}

template < class T > 
Grid< T >::~Grid()
{
  if (width > 0) delete [] grid;
}

Sample::Sample(): 
  h(0), a(sampleFill), range(-1,-1) 
{};

Sample::Sample(AnnotatedPoint p): 
  h(p.h), a(p.a), range(p.range) 
{}; 

Sample::Sample(DOUBLE _h, SampleCategory _a): 
  h(_h), a(_a), range(_h,_h) 
{};
  
DOUBLE Sample::weight() const
{
  return range.second - range.first;
}

DOUBLE Sample::priority() const
{
  return (a == sampleJumpCore ? weight() : 0); 
}

Slab::Slab(): 
  bottom(0), top(0) 
{};

Slab::Slab(DOUBLE _bottom, DOUBLE _top): 
  bottom(_bottom), top(_top) 
{};

void Slab::decideWallAndParapet(
  const SlabList& neighbourSlabs, 
  std::vector< DOUBLE >& wallStartStop, DOUBLE& parapet
) const
{
  // first find highest neighbour ceiling not above current floor
  SlabList::const_reverse_iterator s = neighbourSlabs.rbegin();
  while (s != neighbourSlabs.rend() && s->bottom > top) ++s;

  DOUBLE drop = top;
  if (s != neighbourSlabs.rend()) drop -= s->top;

  if (top == 1.0 || drop < noParapetDrop) parapet = 0;
  else parapet = std::min((DOUBLE) 1.0, drop / maxParapetDrop);
  
  // former check for top > 0 removed (in preparation for sound 3D models)

  if (s == neighbourSlabs.rend())
  {
    wallStartStop.push_back(top);
    wallStartStop.push_back(bottom);
    return;
  }
  if (s->top < top)
  {
    wallStartStop.push_back(top);
    wallStartStop.push_back(max(s->top, bottom));
  }
  while (s->bottom > bottom)
  {
    wallStartStop.push_back(s->bottom);
    ++s;
    if (s == neighbourSlabs.rend() || s->top <= bottom)
    {
      wallStartStop.push_back(bottom);
      return;
    }
    wallStartStop.push_back(s->top);
  }
}

PointCollector& PointCollector::operator<<(const AnnotatedPoint& p)
{
  Complaint(internalError) 
    << "Calling pure virtual function PointCollector::operator<<(...)";
  return *this;
}

bool PointCollector::empty() const
{
  Complaint(internalError) 
    << "Calling pure virtual function PointCollector::empty()";
  return true;
}

AnnotatedPoint PointCollector::back() const
{
  return lastPoint;
}

PointRecorder::PointRecorder(std::vector< Point >& _pp): pp(_pp) {};

PointCollector& PointRecorder::operator<<(const AnnotatedPoint& p)
{
  lastPoint = p;
  pp.push_back((Point) p);
  return *this;
}

bool PointRecorder::empty() const
{
  return pp.empty();
}

void SampleCollector::addSampleToCell(int r, int c, const Sample& s)
{
  Complaint(internalError) 
    << "Calling pure virtual function SampleCollector::addSampleToCell(...)";
}

void SampleCollector::calculateSlabs(int r, int c, SlabList& slabs)
{
  Complaint(internalError) 
    << "Calling pure virtual function SampleCollector::calculateSlabs(...)";
}

void SampleCollector::registerAccess(int r, int c)
{
  if (maxr < minr) { minr = r; maxr = r; minc = c; maxc = c; return; }
  minr = min(minr, r);
  maxr = max(maxr, r);
  minc = min(minc, c);
  maxc = max(maxc, c);
}

SampleCollector::SampleCollector(int _cellsPerUnit):
  minr(0), maxr(-1), minc(0), maxc(-1),
  cellsPerUnit(_cellsPerUnit)
{};
  
PointCollector& SampleCollector::operator<<(const AnnotatedPoint& p)
{
  TrianglePoint< int > t = round(TrianglePoint< DOUBLE >(p * cellsPerUnit));
  Sample s(p);
  // limit elevation to very slightly below 1, to be able to tell
  // the difference with the upper reference plane:
  s.h = std::min((DOUBLE) 0.9999, s.h);
  registerAccess(t[0], t[1]);
  addSampleToCell(t[0], t[1], s);
  lastPoint = p;
  return *this;
}

bool SampleCollector::empty() const
{
  return (maxr < minr);
}

void SampleCollector::calculateSlabs(Grid< SlabList >& slabs, int padding)
{
  int targetMinr = minr - padding;
  int targetMaxr = maxr + padding;
  int targetMinc = minc - padding;
  int targetMaxc = maxc + padding;
  // ensure odd number of rows and columns (even difference between max and min):
  if (targetMaxr - targetMinr & 1) ++targetMaxr;
  if (targetMaxc - targetMinc & 1) ++targetMaxc;

  if (Options::verbosity >= 1) cerr << "Rendering grid size: " 
    << (targetMaxr-targetMinr+1) << " x " << (targetMaxc-targetMinc+1) 
    << "." << endl;
  slabs.grow(targetMinr, targetMaxr, targetMinc, targetMaxc);
  for (int r = targetMinr; r <= targetMaxr; ++r)
  { 
    reportProgress(
      (DOUBLE) (r-targetMinr) / (targetMaxr-targetMinr+1), 
      "Collecting slabs"
    );
    for (int c = targetMinc; c <= targetMaxc; ++c)
      calculateSlabs(r, c, slabs(r,c));
  }
  reportProgress(1, "Collecting slabs");
}

void MultilevelCollector::addSampleToCell(int r, int c, const Sample& s)
{
  CellCollector& cell = grid(r,c);
  if (cell.slabs.empty() || s.h >= cell.slabs.back().top + strength + clearance)
  {
    cell.slabs.push_back(Slab(s.h - strength, s.h));
    cell.priority = s.priority();
  }
  else if (s.priority() >= cell.priority)
  {
    cell.slabs.back().top = s.h;
    cell.priority = s.priority();
  }
  if (s.a == sampleFill) 
    cell.slabs.front().bottom = min(cell.slabs.front().bottom, foundation);
}

void MultilevelCollector::calculateSlabs(int r, int c, SlabList& slabs)
{
  CellCollector& cell = grid(r,c);
  for (SlabList::iterator s = cell.slabs.begin(); s != cell.slabs.end(); ++s)
    slabs.push_back(*s);
}

MultilevelCollector::MultilevelCollector(
  int _cellsPerUnit, DOUBLE _strength, DOUBLE _clearance, DOUBLE _foundation
): strength(_strength), clearance(_clearance), foundation(_foundation),
  SampleCollector(_cellsPerUnit)
{};

MeanCollector::CellCollector::CellCollector(): sum(0), weight(0), count(0) {};

void MeanCollector::addSampleToCell(int r, int c, const Sample& s)
{
  CellCollector& cell = grid(r,c);
  static const DOUBLE practicallyZero = 0.0000001;
  DOUBLE w = s.weight();
  if (w < cell.weight - practicallyZero) return;
  if (w > cell.weight + practicallyZero)
  {
    cell.sum = 0; cell.count = 0; cell.weight = w;
  }
  cell.sum += s.h; ++cell.count;
}

void MeanCollector::calculateSlabs(int r, int c, SlabList& slabs)
{
  CellCollector& cell = grid(r,c);
  if (cell.count == 0) slabs.push_back(Slab(foundation, floorHeight));
  else slabs.push_back(Slab(foundation, cell.sum/cell.count));
}

MeanCollector::MeanCollector(int _cellsPerUnit, DOUBLE _foundation):
  foundation(_foundation), SampleCollector(_cellsPerUnit)
{};

void constructUpperPlane(
  Grid< SlabList >& slabs, DOUBLE foundation, int discreteRidge
)
{
  for (Grid< SlabList >::iterator i = 
    slabs.begin("Constructing upper reference plane"); i(); ++i)
  {
    if (i.location() * Point(1,-1) < discreteRidge) continue;
    if ((*i).empty()) (*i).push_back(Slab(foundation, 1.0));
    else // make make sure the bottom slab extends to the bottom:
      (*i).front().bottom = foundation;
  }
}

ElevatedPoint cameraPosition(
  Grid< SlabList >& slabs,
  const Similarity& ModelCoordFromGridCoord,
  const View& view
)
{
  // first figure out where the camera needs to be to see everything:
  const DOUBLE t = 0.33; // tan of (field of view / 2)
  const DOUBLE degtorad = 3.14159265358979323846 / 180.0;
  const DOUBLE sinazm = sin(view.azimuth * degtorad);
  const DOUBLE cosazm = cos(view.azimuth * degtorad);
  const DOUBLE sinalt = sin(view.altitude * degtorad);
  const DOUBLE cosalt = cos(view.altitude * degtorad);
  DOUBLE topEdgeConstraint = 100;    // t*camera.x - camera.h < topEdgeConstraint
  DOUBLE bottomEdgeConstraint = 100; // t*camera.x + camera.h < bottomEdgeConstraint
  DOUBLE leftEdgeConstraint = 100;   // t*camera.x - camera.y < leftEdgeConstraint
  DOUBLE rightEdgeConstraint = 100;  // t*camera.x + camera.y < rightEdgeConstraint
  for (Grid< SlabList >::iterator i = 
    slabs.visitInterior("Calculating boundaries for camera position"); i(); ++i)
  {
    if ((*i).empty()) continue;
    if ((*i).back().top < 0) continue;
    if ((*i).back().top == 1) 
    {
      // ignore upper reference plane unless it is on the edge with the trailscape
      bool onEdge = false;
      for (int j = 0; !onEdge && j < 6; ++j)
      {
        SlabList& neighbour = i.hexNeighbour(j);
        onEdge = (
          !neighbour.empty() && 
          neighbour.back().top > 0 && 
          neighbour.back().top < 1
        ); 
      }
      if (!onEdge) continue;
    }
    Point p = ModelCoordFromGridCoord(i.location()); 
    ElevatedPoint e[2];
    e[0] = ElevatedPoint(p, (*i).front().bottom);
    e[1] = ElevatedPoint(p, (*i).back().top);
    for (int j = 0; j < 2; ++j)
    {
      // rotate scene so that camera looks into direction of positive x-axis
      ElevatedPoint tmp(
        e[j].x*cosazm + e[j].y*sinazm, -e[j].x*sinazm + e[j].y*cosazm, e[j].h);
      e[j] = ElevatedPoint(
        tmp.x*cosalt - tmp.h*sinalt, tmp.y, tmp.x*sinalt + tmp.h*cosalt);
      DOUBLE tqx = t * e[j].x;
      topEdgeConstraint = std::min(topEdgeConstraint, tqx - e[j].h);
      bottomEdgeConstraint = std::min(bottomEdgeConstraint, tqx + e[j].h);
      leftEdgeConstraint = std::min(leftEdgeConstraint, tqx - e[j].y);
      rightEdgeConstraint = std::min(rightEdgeConstraint, tqx + e[j].y);
    }
  }
  ElevatedPoint camera;
  camera.y = (view.defaultCentre ? 
      (rightEdgeConstraint - leftEdgeConstraint) / 2 : 0);
  camera.h = (view.defaultCentre ? 
      (bottomEdgeConstraint - topEdgeConstraint) / 2 : view.centreHeight);
  camera.x = view.distance * std::min(
    std::min(topEdgeConstraint + camera.h,
    bottomEdgeConstraint - camera.h),
    std::min(leftEdgeConstraint + camera.y,
    rightEdgeConstraint - camera.y)
  ) / t;
  return camera;
}

void smoothen(Grid< SlabList >& slabs, DOUBLE foundation)
{
  Point l = slabs.centre();
  int r = (int) l.x;
  int c = (int) l.y;
  for (Grid< SlabList >::iterator i = slabs.visitInterior(); i(); ++i)
  {
    DOUBLE sum = (*i).front().top;
    for (int j = 0; j < 6; ++j) 
      sum += i.hexNeighbour(j).front().top;
    // abuse bottom field to store new value
    (*i).front().bottom = sum / 7.0;
  }
  // move new values into place
  for (Grid< SlabList >::iterator i = slabs.visitInterior(); i(); ++i)
  {
    Slab& s = (*i).front();
    s.top = s.bottom;
    s.bottom = foundation;
  }
}

#ifdef NARROWS
void widenPassages(Grid< SlabList >& slabs,
  DOUBLE cellSpacing, // horizontal distance between centres of adjacent grid cells,
                      // where 1 is the height of the model (above ground level)
  DOUBLE strength,    // minimum thickness of material
  DOUBLE clearance,   // minimum height of tunnels (= min width of openings?)
  DOUBLE factor       // parameter to control how wide passages should be relative
                      // to their length; values would probably range between 
                      // 0.01 and 0.03 
)
{
/*
  This function should modify the contents of a grid of hexagons, that stores,
  for each cell, a sequence of "slabs" (from bottom to top), where a "slab" is 
  an elevation interval in which solid material should be printed. This information
  is stored in objects of the following classes.
  
  An object x of class Slab models an elevation interval, whose bounds are
  stored in the member variables x.bottom and x.top, of type DOUBLE.
  Creation is through the constructor Slab(bottom,top).

  For each grid cell, its slabs are stored as a std::vector< Slab > (aka SlabList),
  in order from bottom to top.

  These list are arranged in a grid called slabs using the class Grid< SlabList >.
  The SlabList of the cell in row r, column c can be accessed with:

    slabs(r,c)

  To iterate over all grid cells, one may use:

    for (Grid< SlabList >::iterator i = slabs.begin(); i(); ++i)
    {
      // in this block, access the current cell with *i
      // (note: -> operator is not implemented)
    }

  For more method that operate on the Grid, see the definition of class Grid.

  The elevations of the slab tops range from 0 to 1, the elevations of the slab 
  bottoms range from the constant value foundation to sligthly less than 1. 
  The horizontal distance between the centres of adjacent grid cells is given by 
  the variable cellSpacing, where 1 is the length unit of the elevation dimension.
  The cells in the outermost rows and columns of the grid are sentinels that will 
  not be rendered and should not be modified.
*/
  if (Options::verbosity >= 1) cerr << "Widening narrow passages..." << endl;
  Complaint(warning) << "Widening of narrow passages has not been implemented yet.";
}
#endif

#ifdef PRINTABLE
void makePrintable(Grid< SlabList >& slabs,
  DOUBLE cellSpacing, DOUBLE strength, DOUBLE clearance
)
{
/* 
  For an explanation of the relevant variables, see widenPassages 
*/
  if (Options::verbosity >= 1) cerr << "Thickening the 3D model..." << endl;
  Complaint(warning) << "3D-printable models have not been implemented yet.";
}
#endif

Mesh::Mesh(const string& _material): 
  material(_material), nrPending(0) 
{};
  
void Mesh::setMaterial(const string& _material)
{
  material = _material;
}

const std::vector< string >& Mesh::materials()
{
  return theMaterials;
}

void Mesh::addVertex(MeshVertex& v, int normalIndex)
{
  if (v.index < 0) 
  {
    v.index = coordinates.size();
    coordinates.push_back(v.point);
  }
  polygons.push_back(v.index);
  polygons.push_back(normalIndex);
  ++nrPending;
}

void Mesh::addVertex(const Point& p, DOUBLE h, int normalIndex)
{
  polygons.push_back(coordinates.size());
  polygons.push_back(normalIndex);
  coordinates.push_back(ElevatedPoint(p, h));
  ++nrPending;
}

void Mesh::addVertex(DOUBLE x, DOUBLE y, DOUBLE h, int normalIndex)
{
  addVertex(Point(x, y), h, normalIndex);
}

void Mesh::endPolygon()
{
  polygonSizes.push_back(nrPending);
  nrPending = 0;
}

bool Mesh::full() const 
{
  return (coordinates.size() > maxMeshFill);
}

void Mesh::clear() 
{ 
  if (nrPending > 0) 
    Complaint(internalError) << "clear() called on a mesh with an unfinished polygon.";
  cout << "  <geometry id=\"landscape" << theMaterials.size() << "\">" << endl
       << "    <mesh>" << endl
       << "      <source id=\"Coordinates" << theMaterials.size() << "\">" << endl
       << "        <float_array id=\"CoordinatesArray" << theMaterials.size() 
       << "\" count=\"" << coordinates.size() * 3 << "\">" << endl;
  for (int i = 0; i < coordinates.size(); ++i)
  {
    cout << "          " << fixed << setprecision(outputPrecision) << -coordinates[i].y
         << ' '          << fixed << setprecision(outputPrecision) <<  coordinates[i].h
         << ' '          << fixed << setprecision(outputPrecision) << -coordinates[i].x
         << endl;
  }
  cout << "        </float_array>" << endl
       << "        <technique_common>" << endl
       << "          <accessor source=\"#CoordinatesArray" << theMaterials.size() 
       << "\" count=\"" << coordinates.size() << "\" stride=\"3\">" << endl
       << "            <param name=\"X\" type=\"float\" />" << endl
       << "            <param name=\"Y\" type=\"float\" />" << endl
       << "            <param name=\"Z\" type=\"float\" />" << endl
       << "          </accessor>" << endl
       << "        </technique_common>" << endl
       << "      </source>" << endl;
  cout << "      <source id=\"Normals" << theMaterials.size() << "\">" << endl
       << "        <float_array id=\"NormalsArray" << theMaterials.size() 
       << "\" count=\"24\">" << endl
       << "           0.86602540  0 -0.5" << endl
       << "           0           0 -1  " << endl
       << "          -0.86602540  0 -0.5" << endl
       << "          -0.86602540  0  0.5" << endl
       << "           0           0  1  " << endl
       << "           0.86602540  0  0.5" << endl
       << "           0           1  0  " << endl
       << "           0          -1  0  " << endl 
       << "        </float_array>" << endl
       << "        <technique_common>" << endl
       << "          <accessor source=\"#NormalsArray" << theMaterials.size() 
       << "\" count=\"8\" stride=\"3\">" << endl
       << "            <param name=\"X\" type=\"float\" />" << endl
       << "            <param name=\"Y\" type=\"float\" />" << endl
       << "            <param name=\"Z\" type=\"float\" />" << endl
       << "          </accessor>" << endl
       << "        </technique_common>" << endl
       << "      </source>" << endl;
  cout << "      <vertices id=\"Vertices" << theMaterials.size() << "\">" << endl
       << "        <input semantic=\"POSITION\" source=\"#Coordinates" << theMaterials.size() << "\"/>" << endl
       << "      </vertices>" << endl
       << "      <polylist count=\"" << polygonSizes.size() << "\" material=\"SOLID" << theMaterials.size() << "\">" << endl
       << "        <input semantic=\"VERTEX\" source=\"#Vertices" << theMaterials.size() << "\" offset=\"0\"/>" << endl
       << "        <input semantic=\"NORMAL\" source=\"#Normals" << theMaterials.size() << "\" offset=\"1\"/>" << endl
       << "        <vcount>" << endl;
  for (int i = 0; i < polygonSizes.size(); ++i) cout << ' ' << polygonSizes[i];
  cout << endl 
       << "        </vcount>" << endl;
  int offset = 0;
  for (int i = 0; i < polygonSizes.size(); ++i)
  {
    cout << "        <p>";
    for (int j = 0; j < 2*polygonSizes[i]; ++j) cout << ' ' << polygons[offset+j];
    cout << " </p>" << endl;
    offset += 2*polygonSizes[i];
  }
  cout << "      </polylist>" << endl
       << "    </mesh>" << endl
       << "  </geometry>" << endl;
  theMaterials.push_back(material);
  coordinates.clear();
  polygonSizes.clear();
  polygons.clear();
}

Mesh::~Mesh() 
{
  if (!coordinates.empty()) clear();
}

std::vector< string > Mesh::theMaterials;

MeshVertex::MeshVertex(Point p, DOUBLE h):
  point(p, h), index(-1) 
{};

void drawModel(
  Grid< SlabList >& slabs,
  const Similarity& ModelCoordFromGridCoord,
  bool printable, 
  int nrColours, 
  DOUBLE parapetHeight
)
{
  WrapArray< Point, 6 > cellShape;
  for (int i = 0; i < 6; ++i)
  {
    const int* d = gridIndexDifferenceToNeighbour[i];
    cellShape[i]   += Point(d[0],d[1]);
    cellShape[i+1] += Point(d[0],d[1]);
  }
  for (int i = 0; i < 6; ++i)
    cellShape[i] = ModelCoordFromGridCoord.rotation(cellShape[i]) / 3.0;

  Mesh terrain[nrColours];
  string colourNames[nrColours+1];
  for (int i = 0; i < nrColours; ++i)
  {
    stringstream name;
    name << "PathMaterial" << i;
    terrain[i].setMaterial(name.str());
  }
  Mesh parapets("ParapetMaterial");
  Mesh walls("WallMaterial");
  Mesh upperPlane("UpperPlaneMaterial");

  for (Grid< SlabList >::iterator c = 
    slabs.visitInterior("Computing facets of 3D model"); c(); ++c)
  {
    if ((*c).empty()) continue;
    Point p = ModelCoordFromGridCoord(c.location());
    for (SlabList::iterator s = (*c).begin(); s != (*c).end(); ++s)
    { 
      if (s->top < 0 && !printable) continue; // skip underground slabs 
      int colour = (int) (s->top * nrColours);
      if (colour < 0) colour = 0; // just to catch rounding errors 
      Mesh& surface = (colour < nrColours ? terrain[colour] : upperPlane);
      if (surface.full()) surface.clear();          
      if (parapets.full()) parapets.clear();
      if (walls.full()) walls.clear();

      // figure out what walls and parapets are needed:
      WrapArray< DOUBLE, 6 > parapetFraction;
      WrapArray< std::vector< DOUBLE >, 6 > wallStartStop;
      for (int i = 0; i < 6; ++i)
      {
        // parapet is constructed where there is a drop > 0.001; 
        // ideal height between 0 and parapetHeight (for drop >= 0.02),
        // fraction of parapetHeight stored in parapetFraction
        s->decideWallAndParapet(
          c.hexNeighbour(i), wallStartStop[i], parapetFraction[i]
        );
        if (parapetHeight <= 0) parapetFraction[i] = 0;
      }

      // define vertices
      WrapArray< MeshVertex, 6 > surfaceTopOutside;
      WrapArray< MeshVertex, 6 > surfaceTopInside;
      WrapArray< MeshVertex, 6 > parapetTopOutside;
      WrapArray< MeshVertex, 6 > parapetTopInside;
      WrapArray< MeshVertex, 6 > parapetParapet;
      WrapArray< MeshVertex, 6 > wallBottom;
      WrapArray< MeshVertex, 6 > wallTopOutside;
      WrapArray< MeshVertex, 6 > wallParapet;
      for (int i = 0; i < 6; ++i)
      {
        // if parapet top is between two parapets, take max of their ideal heights
        DOUBLE postFraction = max(parapetFraction[i-1], parapetFraction[i]);
        DOUBLE postHeight = postFraction * parapetHeight;
        DOUBLE postDepth = 
          minSurfaceToParapetRatio / (postFraction + minSurfaceToParapetRatio); 
        surfaceTopOutside[i] = MeshVertex(p + cellShape[i], s->top);
        surfaceTopInside[i] = MeshVertex(p + cellShape[i] * postDepth, s->top);
        parapetTopOutside[i] = MeshVertex(p + cellShape[i], s->top);
        parapetTopInside[i] = MeshVertex(p + cellShape[i] * postDepth, s->top);
        parapetParapet[i] = MeshVertex(p + cellShape[i], s->top + postHeight);
        wallBottom[i] = MeshVertex(p + cellShape[i], s->bottom);
        wallTopOutside[i] = MeshVertex(p + cellShape[i], s->top);
        wallParapet[i] = MeshVertex(p + cellShape[i], s->top + postHeight);
      }

      // construct the top
      for (int i = 0; i < 6; ++i)
      {
        if (parapetFraction[i-1] == 0) {
          surface.addVertex(surfaceTopOutside[i], 6);
          if (parapetFraction[i] != 0) surface.addVertex(surfaceTopInside[i], 6);
        } else {
          surface.addVertex(surfaceTopInside[i], 6);
          if (parapetFraction[i] == 0) surface.addVertex(surfaceTopOutside[i], 6);
        }
      }
      surface.endPolygon();

      // construct parapets
      for (int i = 0; i < 6; ++i) {
        if (parapetFraction[i] == 0) continue;            
        if (parapetFraction[i-1] == 0) {
          parapets.addVertex(parapetTopInside[i],  (i+4)%6);
          parapets.addVertex(parapetParapet[i],    (i+4)%6);
          parapets.addVertex(parapetTopOutside[i], (i+4)%6);
          parapets.endPolygon();
        }
        parapets.addVertex(parapetTopInside[i],   (i+3)%6);
        parapets.addVertex(parapetTopInside[i+1], (i+3)%6);
        parapets.addVertex(parapetParapet[i+1],   (i+3)%6);
        parapets.addVertex(parapetParapet[i],     (i+3)%6);
        parapets.endPolygon();
        if (parapetFraction[i+1] == 0) {
          parapets.addVertex(parapetTopInside[i+1],  (i+2)%6);
          parapets.addVertex(parapetParapet[i+1],    (i+2)%6);
          parapets.addVertex(parapetTopOutside[i+1], (i+2)%6);
          parapets.endPolygon();
        }
      }

      // construct the vertical walls
      for (int i = 0; i < 6; ++i)
        for (std::vector< DOUBLE >::const_iterator w = wallStartStop[i].begin();
          w != wallStartStop[i].end(); ++w)
        {
          // top vertices
          if (*w == s->top)
          {
            if (parapetFraction[i] == 0)
            {
              walls.addVertex(wallTopOutside[i+1], i);
              walls.addVertex(wallTopOutside[i],   i);
            } 
            else
            {
              walls.addVertex(wallParapet[i+1], i);
              walls.addVertex(wallParapet[i],   i);
            }
          }
          else
          {
            walls.addVertex(p + cellShape[i+1], *w, i);
            walls.addVertex(p + cellShape[i],   *w, i);
          }

          // bottom vertices
          ++w;
          if (*w == s->bottom)
          {
            walls.addVertex(wallBottom[i],   i);
            walls.addVertex(wallBottom[i+1], i);
          }
          else
          {
            walls.addVertex(p + cellShape[i],   *w, i);
            walls.addVertex(p + cellShape[i+1], *w, i);
          }
          walls.endPolygon();
        }

      // construct the bottom (only if above the reference plane)
      if (printable || s->bottom > 0)
      {
        for (int i = 0; i < 6; ++i) walls.addVertex(wallBottom[i], 7);
        walls.endPolygon();
      }
    }   
  }
  if (Options::verbosity >= 1) 
  {
    cerr << "Finishing COLLADA file...\r";
    cerr.flush();
  } 
}

void drawLowerReferencePlane()
{
  Mesh lowerPlane("LowerPlaneMaterial");
  lowerPlane.addVertex(-100,  100, 0, 6);
  lowerPlane.addVertex( 100,  100, 0, 6);
  lowerPlane.addVertex( 100, -100, 0, 6);
  lowerPlane.addVertex(-100, -100, 0, 6);
  lowerPlane.endPolygon();
}

void drawUpperReferencePlane(
  Grid< SlabList >& slabs,
  const Similarity& ModelCoordFromGridCoord,
  int discreteRidge
)
{
  Point gridCorner[5];
  gridCorner[0] = slabs.corner(0) + Point( 0.67,-0.67);
  gridCorner[1] = slabs.corner(1) + Point( 0.67, 0.67);
  gridCorner[2] = slabs.corner(2) + Point(-0.67, 0.67);
  gridCorner[3] = slabs.corner(3) + Point(-0.67,-0.67);
  for (int i = 0; i < 4; ++i) 
    gridCorner[i] = ModelCoordFromGridCoord(gridCorner[i]);
  gridCorner[4] = gridCorner[0];
  DOUBLE ridgeLine = ModelCoordFromGridCoord(
//    (Point(discreteRidge-1, 0) + Point(discreteRidge, 0)) * 0.5
    Point(discreteRidge, 0)
  ).x;
  int firstCorner = 0;
  int lastCorner = 4;
  if (ridgeLine >= gridCorner[2].x) 
    lastCorner = -1;
  else
  {
    if (ridgeLine >= gridCorner[1].x) ++firstCorner;
    gridCorner[firstCorner].y += 
      (gridCorner[firstCorner+1].y - gridCorner[firstCorner].y) *
      (ridgeLine - gridCorner[firstCorner].x) /
      (gridCorner[firstCorner+1].x - gridCorner[firstCorner].x);
    gridCorner[firstCorner].x = ridgeLine;
    if (ridgeLine >= gridCorner[3].x) --lastCorner;
    gridCorner[lastCorner].y += 
      (gridCorner[lastCorner-1].y - gridCorner[lastCorner].y) *
      (ridgeLine - gridCorner[lastCorner].x) /
      (gridCorner[lastCorner-1].x - gridCorner[lastCorner].x);
    gridCorner[lastCorner].x = ridgeLine;
  }

  Mesh walls("WallMaterial");
  walls.addVertex(ridgeLine, 100, 0, 4); 
  walls.addVertex(ridgeLine, 100, 1, 4);
  if (firstCorner < lastCorner)
  { 
    walls.addVertex(gridCorner[firstCorner], 1, 4); 
    walls.addVertex(gridCorner[firstCorner], 0, 4); 
    walls.endPolygon();
    walls.addVertex(gridCorner[lastCorner], 0, 4); 
    walls.addVertex(gridCorner[lastCorner], 1, 4); 
  }
  walls.addVertex(ridgeLine, -100, 1, 4); 
  walls.addVertex(ridgeLine, -100, 0, 4); 
  walls.endPolygon();

  Mesh upperPlane("UpperPlaneMaterial");
  upperPlane.addVertex(ridgeLine, 100, 1, 6);
  for (int i = firstCorner; i <= lastCorner; ++i)
    upperPlane.addVertex(gridCorner[i], 1, 6);
  upperPlane.addVertex(ridgeLine, -100, 1, 6);
  upperPlane.addVertex(100, -100, 1, 6);
  upperPlane.addVertex(100, 100, 1, 6);
  upperPlane.endPolygon();
}


/*********************************************************
* main                                                   *
*********************************************************/

int main(const int argc, const char* argv[]) {
  if (argc < 2) Options::usage();
  Options options;
  options.parse(argc, argv);

  // ifs.generatePoints guarantees that sample density suffices for grid
  // cells with diameter options.ifs.diameterLowerBound() / resolution, 
  // which means 
  // 2.0 * SQRTTHIRD * resolution / options.ifs.diameterLowerBound()
  // cells per unit in sample space along an edge/row/column of the grid
  DOUBLE cellsPerUnit = 
    SQRTTHIRD * 2.0 * options.resolution / options.ifs.diameterLowerBound();
  SampleCollector* pp;
  if (options.erosionRadius < 0) 
    pp = new MultilevelCollector(
      cellsPerUnit, 
      options.strength,
      (options.supported ? options.clearance : 0),
      (options.supported ? options.foundation : 1)
    );
  else
    pp = new MeanCollector(cellsPerUnit, options.foundation);
  if (options.view.orientation.size() > 1 && Options::verbosity >= 1) 
    cerr << "Sampling " << options.view.orientation.size() 
      << " copies of the traversal:" << endl;
  for (std::vector< Similarity >::const_iterator 
    o = options.view.orientation.begin(); 
    o != options.view.orientation.end(); 
    ++o
  )
    options.ifs.generatePoints(*pp, *o);

  // calculate slabs and add padding for eroded and sentinel border:
  Grid< SlabList > slabs;
  pp->calculateSlabs(slabs, 2 + max(options.erosionRadius, 0));
  delete pp;

  int discreteRidge;
  if (options.view.ridge) 
  {
    // construct upper reference plane.
    // it consists of all cells whose centre points have x > ridgeLine
    // in sample space;
    // these are exactly the cells with row - column >= discreteRidge, 
    // where:
    discreteRidge = (int) (2 * options.view.ridgeLine * cellsPerUnit) + 2;
    constructUpperPlane(slabs, options.foundation, discreteRidge);
  }

  // calculate the mapping from grid coordinates to model coordinates;
  // to ensure consistency, construct it rather than hard-code it.
  // cellSpacing := cell width in model space, measured between parallel sides
  DOUBLE cellSpacing = cornerDepth * 2.0 / (slabs.edgeLength() - 2); 
  Point stepInColumn(((Point) TrianglePoint< int >(0, 1, 0)) * cellSpacing);
  Point stepInRow   (((Point) TrianglePoint< int >(0, 0, 1)) * cellSpacing);
  Rotation shape(
    Point(stepInColumn.x, stepInRow.x),
    Point(stepInColumn.y, stepInRow.y)
  );
  Similarity ModelCoordFromGridCoord(
    shape, 
    ElevatedPoint(
      (options.view.defaultCentre ? -shape(slabs.centre()) : Point()),
      0
    )   
  );
  
  // calculate the camera position; must be done before smoothing because
  // otherwise it will zoom out to see the complete smoothed edge between
  // the lower and the upper reference plane 
  ElevatedPoint camera = 
    cameraPosition(slabs, ModelCoordFromGridCoord, options.view);

  if (options.erosionRadius > 0)
  {
    for (int i = 0; i < options.erosionRadius; ++i) 
    {
      reportProgress((DOUBLE) i / options.erosionRadius, "Smoothening");
      smoothen(slabs, options.foundation);
    }
    reportProgress(1, "Smoothening");
  }

#ifdef NARROWS
  if (options.widePassages) 
    widenPassages(slabs,
      cellSpacing,
      options.strength, 
      (options.supported ? options.clearance : 0),
      options.bridgeWidth
    );
#endif
#ifdef PRINTABLE
  if (options.printable)
    makePrintable(slabs, cellSpacing, options.strength, options.clearance);
#endif

  cout << "<!--" << endl;
  cout << "Path-filling trail model generated by " << pftrailInfo << endl;
  cout << endl;
  options.report(cout);
  options.ifs.report(cout);
  options.view.report(cout);
  cout << "-->" << endl;

  cout << "<library_geometries>" << endl;
  drawModel(
    slabs, 
    ModelCoordFromGridCoord, 
    options.printable, 
    options.nrColourClasses,
    options.parapet
  );
  if (!options.printable) drawLowerReferencePlane();
  if (options.view.ridge) drawUpperReferencePlane(
    slabs,
    ModelCoordFromGridCoord,
    discreteRidge
  );

  cout << "</library_geometries>" << endl << endl;

  cout << "<library_nodes>" << endl;
  cout << "  <node id=\"Landscape\">" << endl;
  for (int i = 0; i < Mesh::materials().size(); ++i)
    cout << "    <node>" << endl
         << "      <instance_geometry url=\"#landscape" << i << "\">" << endl
         << "        <bind_material>" << endl
         << "                <technique_common>" << endl
         << "                  <instance_material symbol=\"SOLID" << i << "\" target=\"#"
         << Mesh::materials()[i] << "\"/>" << endl
         << "                </technique_common>" << endl
         << "        </bind_material>" << endl
         << "      </instance_geometry>" << endl
         << "    </node>" << endl;
  cout << "  </node>" << endl;
  cout << "  <node id=\"HighCamera\">" << endl
       << "    <rotate> 0 1 0 " << options.view.azimuth << " </rotate>" << endl
       << "    <rotate> 1 0 0 " << -options.view.altitude << " </rotate>" << endl
       << "    <translate> "
       << fixed << setprecision(outputPrecision) << -camera.y << ' ' 
       << fixed << setprecision(outputPrecision) <<  camera.h << ' ' 
       << fixed << setprecision(outputPrecision) << -camera.x
       << " </translate>" << endl
       << "    <instance_camera url=\"#TheCamera\" />" << endl
       << "  </node>" << endl
       << "</library_nodes>" << endl;
  if (Options::verbosity >= 1) cerr << "Finishing COLLADA file: done." << endl;
  cout << "<!-- End of pftrail output. Append postamble here to complete the file. -->" << endl;
}


/*********************************************************
**********************************************************
**                                                      **
** HISTORY                                              **
**                                                      **
**********************************************************
**********************************************************

2020-04-28 added feature: enabled multiple render lines in 
           view configuration, to render multiple copies 
           of a curve. KNOWN BUG AT THIS POINT:
           implementation does not generate all samples in
           order from top to bottom, which results in 
           minor asymmetries due to the behaviour of 
           MultilevelCollector::addSampleToCell

2020-04-28 restructured so that IFS.generatePoints sends
           samples to the grid directly for immediate 
           processing, so that superfluous samples are 
           discarded a.s.a.p., saving time and memory

2020-04-27 fixed bug in error message about multiple
           input files on command line

2020-04-27 input is now specified with -i option 
           (but "-i" can still be omitted)

2020-04-22 added framework for options to widen narrow 
           passages (-n) and to make a 3D printable model 
           (-s 3d); the desired functionality is still 
           unimplemented at this point

2020-04-20 added feature: asymmetric jumps

2020-04-17 fixed bug in error message about invalid 
           reflection in segment parameters

2020-03-31 released version 1.01

2020-03-31 fixed strtofstop, strtoistop, and 
           IFS::operator() to ignore trailing whitespace

2020-03-25 released version 1.0

*/

